"""Misc. tools for supporting Layer implementations."""


def pop_shape(keyword_options):
    """Parse first keyword option that begins with a digit as a shape.

    Parameters
    ----------
    keyword_options : list of str.
        Result of `keyword.split('.').

    Returns
    -------
    shape : tuple or None

    Examples
    --------
    >>> keyword = 'test.5'
    >>> pop_shape(keyword.split('.'))
    (5,)

    >>> keyword = 'test.more.1x2x3'
    >>> pop_shape(keyword.split('.'))
    (1,2,3)

    >>> keyword = 'test.3.2.1'
    >>> pop_shape(keyword.split('.'))
    (3,)
    
    """

    for i, op in enumerate(keyword_options):
        if (op[0].isdigit()):
            dims = op.split('x')
            shape = tuple([int(d) for d in dims])
            break
    else:
        shape = None
    
    if shape is not None: _ = keyword_options.pop(i)

    return shape


def require_shape(layer, kwargs, minimum_ndim=0, maximum_ndim=None):
    """Raise `TypeError` if `shape` is not specified."""

    shape = kwargs.get('shape', None)
    layer_name = type(layer).__name__
    message = None

    if not isinstance(shape, (tuple, list)):
        message = f'Layer {layer_name} requires a `shape` kwarg.'
    elif len(shape) < minimum_ndim:
        message = (f'Layer {layer_name}.shape must have at least '
                   f'{minimum_ndim} dimensions.')
    elif (maximum_ndim is not None) and (len(shape) > maximum_ndim):
        message = (f'Layer {layer_name}.shape must have no more than '
                   f'{maximum_ndim} dimensions.')

    if message is not None:
        raise TypeError(message)



"""Cost functions for TensorFlowBackend.

Cost function code by Alexander Tomlinson (Github: arrrobase),
ported from a previous version of NEMS

TODO: Update documentation, revise functions. Lots of hard-coded magic numbers
      and assumptions about data format.

"""

import numpy as np
import tensorflow as tf
import tensorflow.keras

from nems.tools.lookup import FindCallable


# Keras' built-in MSE with default options
keras_mse = tf.keras.losses.MeanSquaredError()

def poisson(response, prediction):
    """Poisson loss."""
    return tf.math.reduce_mean(prediction - response * tf.math.log(prediction + 1e-5), name='poisson')


def drop_nan(response, prediction):
    mask = tf.math.is_finite(response)
    return tf.boolean_mask(response, mask), tf.boolean_mask(prediction, mask)


# TODO: looks like this is also normalizing by... I guess variance of response?
#       (but assumes mean is 0)
def loss_se(response, prediction):
    """Squared error loss."""
    return (tf.math.reduce_mean(tf.math.square(response - prediction))) / (tf.math.reduce_mean(tf.math.square(response)))


def loss_tf_nmse_shrinkage(response, prediction):
    """Normalized means squared error with shrinkage loss."""
    return tf_nmse_shrinkage(response, prediction)


def loss_tf_nmse(response, prediction, per_cell=True):
    """Normalized means squared error loss."""
    mE, sE = tf_nmse(response, prediction, per_cell=per_cell)
    if per_cell:
        return tf.math.reduce_mean(mE)
        #return tf.reduce_mean(tf.boolean_mask(mE, tf.math.is_finite(mE)))
    else:
        return mE


def tf_nmse_shrinkage(response, prediction, shrink_factor=0.5, per_cell=True, thresh=False):
    """Calculates the normalized mean squared error, with an adjustment for error.
    Averages across the batches, but optionally can return a per cell error.
    :param response:
    :param prediction:
    :param float shrink_factor:
    :param bool per_cell: Whether to also average over cells or not
    :param bool thresh:
    :return: a "shrunk" normalized mean squared error
    """
    mE, sE = tf_nmse(response, prediction, per_cell)

    def shrink(mE, sE, shrink_factor, thresh):
        def shrink_all(mE, sE, shrink_factor, thresh):
            return tf_shrinkage(mE, sE, shrink_factor, thresh)

        def shrink_some(mE, sE, shrink_factor, thresh):
            mask_gt, mask_lt = mE >= 1, mE < 1
            # make zero where mE was > 1
            shrunk = tf_shrinkage(mE, sE, shrink_factor, thresh) * tf.dtypes.cast(mask_lt, mE.dtype)
            # add back in
            mE = shrunk + mE * tf.dtypes.cast(mask_gt, mE.dtype)
            return mE

        mE = tf.cond(tf.math.reduce_all(mE < 1), lambda: shrink_all(mE, sE, shrink_factor, thresh),
                     lambda: shrink_some(mE, sE, shrink_factor, thresh))
        return mE

    mE = tf.cond(tf.math.reduce_any(mE < 1), lambda: shrink(mE, sE, shrink_factor, thresh), lambda: mE)

    if per_cell:
        mE = tf.math.reduce_mean(mE)

    return mE


def tf_nmse(response, prediction, per_cell=True):
    """Calculates the normalized mean squared error across batches.
    Optionally can return an average per cell.
    :param response:
    :param prediction:
    :param per_cell: Whether to average across all cells or not
    :return: 2 tensors, one of the mean error, the other of the std of the error. If not per cell, then
     tensor is of shape (), else tensor if of shape (n_cells,) (i.e. last dimension of the resp/pred tensor)
    """
    # hardcoded to use 10 jackknifes for error estimate
    n_drop = response.get_shape().as_list()[1] % 10
    if n_drop:
        # use slices to handle varying tensor shapes
        drop_slice = [slice(None) for i in range(len(response.shape))]

        # second dim is time
        drop_slice[1] = slice(None, -n_drop)
        drop_slice = tuple(drop_slice)

        _response = response[drop_slice]
        _prediction = prediction[drop_slice]
    else:
        _response = response
        _prediction = prediction
    print("In tf_nmse:" , _response.shape, _prediction.shape, 'n_drop:', n_drop)

    if per_cell:
        # Put last dimension (number of output channels) first.
        _response = tf.experimental.numpy.moveaxis(_response, [-1, 1], [0, 1])
        _prediction = tf.experimental.numpy.moveaxis(_prediction, [-1, 1], [0, 1])
        
        _response = tf.reshape(_response, shape=(_response.shape[0], 10, -1))
        _prediction = tf.reshape(_prediction, shape=(_prediction.shape[0], 10, -1))
    else:
        _response = tf.reshape(_response, shape=(10, -1))
        _prediction = tf.reshape(_prediction, shape=(10, -1))

    squared_error = ((_response - _prediction) ** 2)
    numers = tf.math.reduce_mean(squared_error, axis=-1)
    denoms = tf.math.reduce_mean(_response**2, axis=-1)
    denoms = tf.where(tf.equal(denoms, 0), tf.ones_like(denoms), denoms)
    
    nmses = (numers / denoms) ** 0.5

    mE = tf.math.reduce_mean(nmses, axis=-1)
    sE = tf.math.reduce_std(nmses, axis=-1) / 10 ** 0.5

    return mE, sE


def tf_shrinkage(mE, sE, shrink_factor=0.5, thresh=False):
    """Adjusts the mean error based on the standard error"""
    mE = 1 - mE
    smd = tf.math.divide_no_nan(abs(mE), sE) / shrink_factor
    smd = 1 - smd ** -2

    if thresh:
        return 1 - mE * tf.dtypes.cast(smd > 1, mE.dtype)

    smd = smd * tf.dtypes.cast(smd > 0, smd.dtype)

    return 1 - mE * smd


# correlation for monitoring
def pearson(y_true, y_pred):
    x = y_true
    y = y_pred
    mx = tf.reduce_mean(x, axis=-2, keepdims=True)
    my = tf.reduce_mean(y, axis=-2, keepdims=True)
    xm, ym = x - mx, y - my
    t1_norm = tf.nn.l2_normalize(xm, axis = -2)
    t2_norm = tf.nn.l2_normalize(ym, axis = -2)
    r = tf.reduce_mean(tf.reduce_sum(t1_norm*t2_norm, axis=[-2], keepdims=True))

    return r


cost_nicknames = {'squared_error': loss_se, 'nmse': loss_tf_nmse,
                  'nmse_shrinkage': loss_tf_nmse_shrinkage}
get_cost = FindCallable({**globals(), **cost_nicknames}, header='Cost')



import numpy as np
import numexpr as ne

from nems.registry import layer
from nems.distributions import Normal
from .base import Layer, Phi, Parameter
from .tools import require_shape, pop_shape


class StateGain(Layer):
    
    state_arg = 'state'  # see Layer docs for details

    def __init__(self, **kwargs):
        """Docs TODO.
        
        Parameters
        ----------
        shape : 2-tuple of int.
            (size of last dimension of state, size of last dimension of input)

        Examples
        --------
        TODO
        
        """
        require_shape(self, kwargs, minimum_ndim=2)
        super().__init__(**kwargs)

    def initial_parameters(self):
        """Docs TODO
        
        Layer parameters
        ----------------
        gain : TODO
            prior:
            bounds:
        offset : TODO
            prior:
            bounds:
        
        """
        zero = np.zeros(shape=self.shape)
        one = np.ones(shape=self.shape)

        gain_mean = zero.copy()
        gain_mean[0,:] = 1  # TODO: Explain purpose of this?
        gain_sd = one/20
        gain_prior = Normal(gain_mean, gain_sd)
        gain = Parameter('gain', shape=self.shape, prior=gain_prior)

        offset_mean = zero
        offset_sd = one
        offset_prior = Normal(offset_mean, offset_sd)
        offset = Parameter('offset', shape=self.shape, prior=offset_prior)
        
        return Phi(gain, offset)

    def evaluate(self, input, state):
        """Multiply and shift input(s) by weighted sums of state channels.
        
        Parameters
        ----------
        input : ndarray
            Data to be modulated by state, typically the output of a previous
            Layer.
        state : ndarray
            State data to modulate input with.

        """

        gain, offset = self.get_parameter_values()
        return np.matmul(state, gain) * input + np.matmul(state, offset)

    @layer('stategain')
    def from_keyword(keyword):
        """Construct StateGain from keyword.
        
        Keyword options
        ---------------
        {digit}x{digit} : specifies shape, (n state channels, n stim channels)
            n stim channels can also be 1, in which case the same weighted
            channel will be broadcast to all stim channels (if there is more
            than 1).
        
        See also
        --------
        Layer.from_keyword

        """
        # TODO: other options from old NEMS
        options = keyword.split('.')
        shape = pop_shape(options)

        return StateGain(shape=shape)

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf.layer_tools import NemsKerasLayer

        class StateGainTF(NemsKerasLayer):

            def call(self, inputs):
                # Assume inputs is a list of two tensors, with state second.
                # TODO: Use tensor names to not require this arbitrary order.
                input = inputs[0]
                state = inputs[1]

                with_gain = tf.multiply(tf.matmul(state, self.gain), input)
                with_offset = with_gain + tf.matmul(state, self.offset)
                
                return with_offset

        return StateGainTF(self, **kwargs)


# Author: SVD, 2022-09-15.
class StateDexp(Layer):
    """State-dependent modulation through double exponential sigmoid."""

    state_arg = 'state'  # see Layer docs for details

    def __init__(self, per_channel=False, **kwargs):
        """Docs TODO.

        Parameters
        ----------
        shape : 2-tuple of int. (state_chans, input_chans), ie,
            (size of last dimension of state, size of last dimension of input)

        Examples
        --------
        TODO

        """
        require_shape(self, kwargs, minimum_ndim=2)
        self.per_channel = per_channel
        super().__init__(**kwargs)

    def initial_parameters(self):
        """Docs TODO

        Layer parameters
        ----------------
        base_g, amp_g, kappa_g, offset_g : TODO
            prior:
            bounds:
        base_d, amp_d, kappa_d, offset_d : TODO
            prior:
            bounds:

        """
        zeros = np.zeros(shape=self.shape)
        ones = np.ones(shape=self.shape)

        # init gain params
        base_mean_g = zeros.copy()
        base_sd_g = ones.copy()
        amp_mean_g = zeros.copy() + 0
        amp_mean_g[:, 0] = 1
        amp_sd_g = ones.copy() * 0.1
        kappa_mean_g = zeros.copy()
        kappa_sd_g = ones.copy() * 0.1
        offset_mean_g = zeros.copy()
        offset_sd_g = ones.copy() * 0.1

        # init dc params
        base_mean_d = zeros.copy()
        base_sd_d = ones.copy()
        amp_mean_d = zeros.copy() + 0
        amp_sd_d = ones.copy() * 0.1
        kappa_mean_d = zeros.copy()
        kappa_sd_d = ones.copy() * 0.1
        offset_mean_d = zeros.copy()
        offset_sd_d = ones.copy() * 0.1

        return Phi(
            Parameter('base_g', shape=self.shape,
                      prior=Normal(base_mean_g, base_sd_g), bounds=(0,10)),
            Parameter('amp_g', shape=self.shape,
                      prior=Normal(amp_mean_g, amp_sd_g), bounds=(0,10)),
            Parameter('kappa_g', shape=self.shape,
                      prior=Normal(kappa_mean_g, kappa_sd_g),
                      bounds=(-np.inf,np.inf)),
            Parameter('offset_g', shape=self.shape,
                      prior=Normal(offset_mean_g, offset_sd_g),
                      bounds=(-np.inf,np.inf)),
            Parameter('base_d', shape=self.shape,
                      prior=Normal(base_mean_d, base_sd_d), bounds=(-10,10)),
            Parameter('amp_d', shape=self.shape,
                      prior=Normal(amp_mean_d, amp_sd_d), bounds=(-10,10)),
            Parameter('kappa_d', shape=self.shape,
                      prior=Normal(kappa_mean_d, kappa_sd_d),
                      bounds=(-np.inf,np.inf)),
            Parameter('offset_d', shape=self.shape,
                      prior=Normal(offset_mean_d, offset_sd_d),
                      bounds=(-np.inf,np.inf))
        )

    def evaluate(self, input, state):
        """Modulate input using double exponential transformation of state.

        TODO: Give some more context here. What's the reasoning for this
              implementation? Relevant publications?
              (Same goes for other Layers).

        Parameters
        ----------
        input : ndarray
            Data to be modulated by state, typically the output of a previous
            Layer.
        state : ndarray
            State data to modulate input with.

        Returns
        -------
        np.ndarray

        See also
        --------
        DoubleExponential.nonlinearity

        """

        base_g, amp_g, kappa_g, offset_g, base_d, amp_d, kappa_d, offset_d = \
            self.get_parameter_values()
        _g = [base_g, amp_g, kappa_g, offset_g]
        _d = [base_d, amp_d, kappa_d, offset_d]
        dexp = "b + a*exp(-exp(-exp(k)*(s-o)))"

        n_states = base_g.shape[0]
        n_inputs = base_g.shape[1]
        state = state[..., :n_states]

        if self.per_channel:
            if input.shape[-1] > 1:
                raise ValueError(
                    "StateDexp per-channel option only supports 1-channel input."
                )
            expression = dexp
            sg = np.empty_like(state)
            sd = np.empty_like(state)

            for i in range(n_states):
                s = state[..., i]
                for parameter_set, array in [(_g, sg), (_d, sd)]:
                    b, a, o, k = [p[i, [0]] for p in parameter_set]
                    s_dexp = ne.evaluate(expression, out=array[..., i])

        else:
            expression = f"sum({dexp}, 1)"
            sg = np.empty_like(input)
            sd = np.empty_like(input)

            for i in range(n_inputs):
                s = state
                for parameter_set, array in [(_g, sg), (_d, sd)]:
                    b, a, o, k = [p[:, i] for p in parameter_set]
                    s_dexp = ne.evaluate(expression, out=array[..., i])

        return sg * input + sd


    @layer('sdexp')
    def from_keyword(keyword):
        """Construct sdexp from keyword.

        Keyword options
        ---------------
        {digit}x{digit} : specifies shape, (n state channels, n stim channels)
            n stim channels can also be 1, in which case the same weighted
            channel will be broadcast to all stim channels (if there is more
            than 1).

        See also
        --------
        Layer.from_keyword

        """

        # TODO: other options from old NEMS
        options = keyword.split('.')
        shape = None
        for op in options[1:]:
            if op[0].isdigit():
                dims = op.split('x')
                shape = tuple([int(d) for d in dims])

        return StateDexp(shape=shape)

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf.layer_tools import NemsKerasLayer

        class SDexpTF(NemsKerasLayer):

            def call(self, inputs):
                # Assume inputs is a list of two tensors, with state second.
                # TODO: Use tensor names to not require this arbitrary order.
                input = inputs[0]
                state = inputs[1]

                bg = tf.expand_dims(tf.expand_dims(self.base_g, 0) ,0)
                ag = tf.expand_dims(tf.expand_dims(self.amp_g, 0) ,0)
                kg = tf.expand_dims(tf.expand_dims(self.kappa_g, 0) ,0)
                og = tf.expand_dims(tf.expand_dims(self.offset_g, 0) ,0)
                state_g = tf.reduce_sum(bg + ag * tf.math.exp(-tf.math.exp(
                    -tf.math.exp(kg) * (tf.expand_dims(state, -1) - og)
                    )), axis=2)
                bd = tf.expand_dims(tf.expand_dims(self.base_d, 0) ,0)
                ad = tf.expand_dims(tf.expand_dims(self.amp_d, 0) ,0)
                kd = tf.expand_dims(tf.expand_dims(self.kappa_d, 0) ,0)
                od = tf.expand_dims(tf.expand_dims(self.offset_d, 0) ,0)
                state_d = tf.reduce_sum(bd + ad * tf.math.exp(-tf.math.exp(
                    -tf.math.exp(kd) * (tf.expand_dims(state, -1) - od)
                    )), axis=2)

                return state_d + state_g * input

        return SDexpTF(self, **kwargs)



import warnings

import numpy as np


def minmax(x, floor=1e-6, warn_on_nan=True, inplace=False):
    """Normalize array `x` within each output channel to have min 0 and max 1.
    
    For array `x` with `x.ndim == m` and `x.shape == (d_1, d_2, ..., d_m)`, the
    first `m-1` dimensions of `x` will be flattened. Minimum and maximum are
    then determined separately for each of the `d_m` output channels, and `x`
    is normalized according to: `y = (x - min(x)) / max(x - min)`.

    In addition to the normalized array, the per-channel minimum and maximum
    values are returned so that the above normalization can be reversed using
    `x = (y*max) + min`.

    In addition to adjusting the minimum and maximum values, 'small' values are
    set to 0. This is helpful for preventing models from fitting to unimportant
    variations due to recording noise in silent/null padding, for example.
    This portion of `minmax` is not reversible. However, an un-normalized array
    `z` will still satisfy `np.allclose(x, z, atol=e)` for `e >= floor*max(max)`.
    See `floor` parameter for details on usage.

    Parameters
    ----------
    x : np.ndarray.
        Data to be normalized. Shape can be:
        (T,)  or  (T, ..., N)  or  (S, T, ..., N)
        where T is the number of time bins, N is the number of output channels
        or neurons, and S is the number of samples or trials. Shape (T,) is
        interpreted as (T,1).
    floor : float; default=1e-6.
        Values in the normalized array (i.e. already mapped to min 0, max 1)
        will be set to 0 if they are less than this value. If you do not wish
        to zero out any values, use `floor=0`.
        NOTE: This operation is not reversible.
    warn_on_nan : bool; default=True.
        If True, raises RuntimeWarning if `x` contains any NaN values. This is
        done so that unexpected NaNs are not silently ignored, but intentional
        NaNs can be handled (i.e. missing data or NaN-padded ragged trials).
    inplace : bool; default=False.
        If True, normalized values are assigned to `x` in-place instead of
        returning a copy. This option helps reduce memory usage for large
        arrays. This is *not* equivalent to `x = minmax(x, inplace=False)`,
        which still tracks a separate copy of `x` within the scope of the
        `minmax` function.

    Returns
    -------
    normalized : np.ndarray.
        Same shape as `x` with min 0 and max 1.
    min_per_channel : np.ndarray.
        Shape (N,), where N is the number of output channels.
    max_per_channel : np.ndarray.
        Shape (N,).

    Raises
    ------
    AttributeError, if flattening the leading axes of `x` would result in a
    copy. See docs for `numpy.reshape`.

    See also
    --------
    undo_minmax

    """

    # TODO: This is my attempt at a compromise between raising an error
    #       (e.g. using min instead of nanmin) and doing nothing. We want to be
    #       able to handle NaNs for ragged trials, for example, but in general
    #       I'd prefer not to let NaNs silently pass through when they're not
    #       expected, since that makes figuring out where they originated much
    #       more difficult.
    if warn_on_nan and (np.isnan(x).sum() > 0):
        warnings.warn(
            "Input array for `normalization.minmax` contains NaN values.",
            RuntimeWarning, stacklevel=2
            )

    # For large arrays, `inplace=True` should reduce memory usage by half.
    if not inplace:
        x = x.copy()

    original_shape = x.shape
    if x.ndim == 1:
        # Make sure there's an output dimension.
        x = x[..., np.newaxis]
    new_shape = (x[..., 0].size, x.shape[-1])
    # Use this instead of reshape to raise error on copy (only want one copy or
    # zero for the entire scope of the function).
    x.shape = new_shape

    # Shift minimum to 0
    min_per_channel = np.nanmin(x, axis=0, keepdims=True)
    min_zero = np.subtract(x, min_per_channel, out=x)

    # Scale max to 1
    max_per_channel = np.nanmax(min_zero, axis=0, keepdims=True)
    # This can only happen if the whole channel is 0, set to 1 just to avoid
    # 0-division error.
    max_per_channel[max_per_channel == 0] = 1
    try:
        normalized = np.divide(min_zero, max_per_channel, out=x)
    except np.core._exceptions.UFuncTypeError as e:
        # Print info about float dtype
        print('Numpy division resulted in a casting error, make sure the '
              'dtype of the input to `normalization.minmax` is float-like.')
        raise e

    # Force data to be exactly 0 for "small" values.
    normalized[normalized < floor] = 0
    normalized.shape = original_shape

    return normalized, min_per_channel, max_per_channel


def undo_minmax(y, _min, _max):
    """Inverse of minmax normalization, `undo_minmax(minmax(x)) == x`.
    
    Parameters
    ----------
    y, _min, _max : (np.ndarray, np.ndarray, np.ndarray)
        As returned by `minmax`.

    Notes
    -----
    Exact equality is only true for `floor = 0` in `minmax`. Even then, there
    could still be small differences due to floating point rounding errors.

    See also
    --------
    minmax
    
    """
    z = y*_max + _min
    z.shape = y.shape
    return z


def joint_minmax(*arrays, **minmax_kwargs):
    """Normalize to min=0, max=1 within channels for all arrays simultaneously.

    TODO: docs.
    NOTE: inplace not supported since arrays are concatenated.
    NOTE: currenly only supports arrays with all dims except 0 same length
    TODO: support different number of output channels.

    """
    # Concatenate in time
    joined = np.concatenate(arrays, axis=0)
    # Minimize as one array
    all_normalized, _min, _max = minmax(joined, inplace=False, **minmax_kwargs)
    # Separate out original arrays
    lengths = [a.shape[0] for a in arrays]
    split_indices = np.cumsum(lengths)
    separate_normalized = np.split(all_normalized, split_indices[:-1], axis=0)

    return separate_normalized



import numpy as np

from nems.registry import layer
from nems.distributions import Normal, HalfNormal
from .base import Layer, Phi, Parameter, ShapeError
from .tools import require_shape, pop_shape


# TODO: double check all shape references after dealing w/ data order etc,
#       make sure correct dims are lined up.
class WeightChannels(Layer):
    """Compute linear weighting of input channels, akin to a dense layer.

    Parameters
    ----------
    shape : N-tuple (usually N=2)
        Determines the shape of `WeightChannels.coefficients`.
        First dimension should match the spectral dimension of the input,
        second dimension should match the spectral dimension of the output.
        Note that higher-dimesional shapes are also allowed and should work
        as-is for this base class, but overall Layer design is intended for
        2-dimensional data so subclasses might not support other shapes.

    See also
    --------
    nems.layers.base.Layer

    Examples
    --------
    >>> wc = WeightChannels(shape=(18,4))
    >>> spectrogram = np.random.rand(10000, 18)  # (time, channels)
    >>> out = spectrogram @ wc.coefficients      # wc.evaluate(spectrogram)
    >>> out.shape
    (10000, 1)

    """
    def __init__(self, **kwargs):
        require_shape(self, kwargs, minimum_ndim=2)
        super().__init__(**kwargs)

    def initial_parameters(self):
        """Get initial values for `WeightChannels.parameters`.
        
        Layer parameters
        ----------------
        coefficients : ndarray
            Shape matches `WeightChannels.shape`.
            Prior:  TODO, currently using defaults
            Bounds: TODO

        Returns
        -------
        nems.layers.base.Phi

        """
        # Mean and sd for priors were chosen mostly arbitrarily, to start
        # with most weights near zero (but not exactly at 0).
        mean = np.full(shape=self.shape, fill_value=0.01)
        sd = np.full(shape=self.shape, fill_value=0.05)
        prior = Normal(mean, sd)

        coefficients = Parameter(
            name='coefficients', shape=self.shape, prior=prior
            )
        return Phi(coefficients)

    @property
    def coefficients(self):
        """Weighting matrix that will be applied to input.
        
        Re-parameterized subclasses should overwrite this so that `evaluate`
        doesn't need to change.

        Returns
        -------
        coefficients : ndarray
            coefficients.shape = WeightChannels.shape
        
        """
        return self.parameters['coefficients'].values

    def evaluate(self, input):
        """Multiply input by WeightChannels.coefficients.

        Computes $y = XA$ for input $X$, where $A$ is
        `WeightChannels.coefficients` and $y$ is the output.
        
        Parameters
        ----------
        input : np.ndarray

        Returns
        -------
        np.ndarray
        
        """

        try:
            output = np.tensordot(input, self.coefficients, axes=(1, 0))
        except ValueError as e:
            # Check for dimension swap, to give more informative error message.
            if 'shape-mismatch for sum' in str(e):
                raise ShapeError(self, input=input.shape,
                                 coefficients=self.coefficients.shape)
            else:
                # Otherwise let the original error through.
                raise e

        return output

    @layer('wc')
    def from_keyword(keyword):
        """Construct WeightChannels (or subclass) from keyword.

        Keyword options
        ---------------
        {digit}x{digit}x ... x{digit} : N-dimensional shape.
        g : Use gaussian function(s) to determine coefficients.

        See also
        --------
        Layer.from_keyword
        
        """
        wc_class = WeightChannels
        kwargs = {}
        options = keyword.split('.')
        kwargs['shape'] = pop_shape(options)

        for op in options:
            if op == 'g':
                wc_class = WeightChannelsGaussian
            elif op == 'b':
                wc_class = WeightChannelsMulti

        wc = wc_class(**kwargs)

        return wc

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs
        
        NOTE: This is currently hard-coded to dot the 2nd dim of the input
        (batch, time, channels, ...) and first dim of coefficients
        (channels, rank, ...).
        
        """
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        class WeightChannelsTF(NemsKerasLayer):
            @tf.function
            def call(self, inputs):
                out = tf.tensordot(inputs, self.coefficients, axes=[[2], [0]])
                return out
        
        return WeightChannelsTF(self, **kwargs)

    @property
    def plot_kwargs(self):
        """Add incremented labels to each output channel for plot legend.
        
        See also
        --------
        Layer.plot
        
        """
        kwargs = {
            'label': [f'Channel {i}' for i in range(self.shape[1])]
        }
        return kwargs

    @property
    def plot_options(self):
        """Add legend at right of plot, with default formatting.

        Notes
        -----
        The legend will grow quite large if there are many output channels,
        but for common use cases (< 10) this should not be an issue. If needed,
        increase figsize to accomodate the labels.

        See also
        --------
        Layer.plot
        
        """
        return {'legend': False}


class WeightChannelsMulti(WeightChannels):
    """WeightChannels specialized for multichannel input (eg, binaural).

    Parameters
    ----------
    shape : N-tuple (usually N=3)

    See also
    --------
    nems.layers.base.Layer
    nems.layers.weight_channels.WeightChannels

    Examples
    --------
    >>> wc = WeightChannelsMulti(shape=(18,4,2))
    >>> spectrogram = np.random.rand(1000, 18, 2)  # (time, spectral_channels, ear)
    >>> # wc.evaluate(spectrogram) equivalent...
    >>> out = np.empty((spectrogram.shape[0], wc.shape[1], wc.shape[2]))
    >>> for i in range(wc.shape[-1]):
            out[:,:,i] = spectrogram[:,:,i] @ wc.coefficients[:,:,i]
    >>> out.shape
    (1000, 4, 2)

    """

    def evaluate(self, input):
        """Multiply `input[...,i]` by `WeightChannels.coefficients[...,i]`."""

        try:
            output = np.moveaxis(np.matmul(np.moveaxis(input, [0,1], [-2,-1]), 
                                           np.moveaxis(self.coefficients, [0,1], [-2,-1])),
                                 [-2,-1], [0, 1])
        except ValueError as e:
            # Check for dimension swap, to give more informative error message.
            if 'mismatch in core dimension' in str(e):
                raise ShapeError(self, input=input.shape,
                                 coefficients=self.coefficients.shape)
            else:
                # Otherwise let the original error through.
                raise e

        return output

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        class WeightChannelsMultiTF(NemsKerasLayer):
            @tf.function
            def call(self, inputs):
                # reshape inputs and coefficients so that mult can happen on last
                # two dimensions. Broadcasting seems to work fine this way
                out = tf.experimental.numpy.moveaxis(
                    tf.matmul(
                        tf.experimental.numpy.moveaxis(inputs, [1, 2], [-2, -1]),
                        tf.experimental.numpy.moveaxis(self.coefficients, [0, 1], [-2, -1])
                        ),
                    [-2, -1], [1, 2]
                    )
                return out

        return WeightChannelsMultiTF(self, **kwargs)


class WeightChannelsGaussian(WeightChannels):
    """As WeightChannels, but sample coefficients from gaussian functions."""

    def initial_parameters(self):
        """Get initial values for `WeightChannelsGaussian.parameters`.

        Layer parameters
        ----------------
        mean : scalar or ndarray
            Mean of gaussian, shape is `self.shape[1:]`.
            Prior:  TODO  # Currently using defaults
            Bounds: (0, 1)
        sd : scalar or ndarray
            Standard deviation of gaussian, shape is `self.shape[1:]`.
            Prior:  TODO  # Currently using defaults
            Bounds: (0*, np.inf)
            * Actually set to machine epsilon to avoid division by zero.

        Returns
        -------
        nems.layers.base.Phi

        """

        mean_bounds = (0, 1)
        sd_bounds = (0, np.inf)

        rank = self.shape[1]
        other_dims = self.shape[2:]
        shape = (rank,) + other_dims
        # Pick means so that the centers of the gaussians are spread across the 
        # available frequencies.
        channels = np.arange(rank + 1)[1:]
        tiled_means = channels / (rank*2 + 2) + 0.25
        for dim in other_dims:
            # Repeat tiled gaussian structure for other dimensions.
            tiled_means = tiled_means[...,np.newaxis].repeat(dim, axis=-1)

        mean_prior = Normal(tiled_means, np.full_like(tiled_means, 0.2))
        sd_prior = HalfNormal(np.full_like(tiled_means, 0.4))
            
        parameters = Phi(
            Parameter(name='mean', shape=shape, bounds=mean_bounds,
                        prior=mean_prior),
            Parameter(name='sd', shape=shape, bounds=sd_bounds,
                        prior=sd_prior, zero_to_epsilon=True)
            )

        return parameters

    @property
    def coefficients(self):
        """Return N discrete gaussians with T bins, where `shape=(T,N)`."""
        mean = self.parameters['mean'].values
        sd = self.parameters['sd'].values
        n_input_channels = self.shape[0]

        x = np.arange(n_input_channels)/n_input_channels
        mean = np.asanyarray(mean)[..., np.newaxis]
        sd = np.asanyarray(sd)[..., np.newaxis]
        coefficients = np.exp(-0.5*((x-mean)/sd)**2)  # (rank, ..., outputs, T)
        reordered = np.moveaxis(coefficients, -1, 0)  # (T, rank, ..., outputs)
        # Normalize by the cumulative sum for each channel
        cumulative_sum = np.sum(reordered, axis=-1, keepdims=True)
        # If all coefficients for a channel are 0, skip normalization
        cumulative_sum[cumulative_sum == 0] = 1
        normalized = reordered/cumulative_sum

        return normalized

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        # TODO: Ask SVD about this kludge in old NEMS code. Is this still needed?
        # If so, explain: I think this was to keep gradient from "blowing up"?
        # Scale up sd bound
        sd, = self.get_parameter_values('sd')
        sd_lower, sd_upper = self.parameters['sd'].bounds
        new_values = {'sd': sd*10}
        new_bounds = {'sd': (sd_lower, sd_upper*10)}

        class WeightChannelsGaussianTF(NemsKerasLayer):
            def call(self, inputs):
                mean = tf.expand_dims(self.mean, -1)
                sd = tf.expand_dims(self.sd/10, -1)
                input_features = tf.cast(tf.shape(inputs)[-1],
                                         dtype=inputs.dtype)
                temp = tf.range(input_features) / input_features
                temp = tf.transpose((temp-mean)/sd, [1,0])
                temp = tf.math.exp(-0.5 * tf.math.square(temp))
                norm = tf.math.reduce_sum(temp, axis=-1, keepdims=True)
                kernel = temp / norm

                return tf.tensordot(inputs, kernel, axes=[[2], [0]])
            
            def weights_to_values(self):
                values = self.parameter_values
                # Undo scaling.
                values['sd'] = values['sd'] / 10
                return values

        return WeightChannelsGaussianTF(self, new_values=new_values,
                                        new_bounds=new_bounds, **kwargs)

    
class WeightChannelsMultiGaussian(WeightChannels):
    """WeightChannels specialized for multichannel input (eg, binaural) and with Gaussian tuning.

    TODO : actually make this work -- current pasted together fomr WeightChannelsMulti and WeightChannelsGaussian, but not tested

    Parameters
    ----------
    shape : N-tuple (usually N=3)

    See also
    --------
    nems.layers.base.Layer
    nems.layers.weight_channels.WeightChannels

    Examples
    --------
    >>> wc = WeightChannelsMultiGaussian(shape=(18,4,2))
    >>> spectrogram = np.random.rand(1000, 18, 2)  # (time, spectral_channels, ear)
    >>> # wc.evaluate(spectrogram) equivalent...
    >>> out = np.empty((spectrogram.shape[0], wc.shape[1], wc.shape[2]))
    >>> for i in range(wc.shape[-1]):
            out[:,:,i] = spectrogram[:,:,i] @ wc.coefficients[:,:,i]
    >>> out.shape
    (1000, 4, 2)

    """

    def initial_parameters(self):
        """Get initial values for `WeightChannelsGaussian.parameters`.

        Layer parameters
        ----------------
        mean : scalar or ndarray
            Mean of gaussian, shape is `self.shape[1:]`.
            Prior:  TODO  # Currently using defaults
            Bounds: (0, 1)
        sd : scalar or ndarray
            Standard deviation of gaussian, shape is `self.shape[1:]`.
            Prior:  TODO  # Currently using defaults
            Bounds: (0*, np.inf)
            * Actually set to machine epsilon to avoid division by zero.

        Returns
        -------
        nems.layers.base.Phi

        """

        mean_bounds = (0, 1)
        sd_bounds = (0, np.inf)

        rank = self.shape[1]
        other_dims = self.shape[2:]
        shape = (rank,) + other_dims
        # Pick means so that the centers of the gaussians are spread across the 
        # available frequencies.
        channels = np.arange(rank + 1)[1:]
        tiled_means = channels / (rank*2 + 2) + 0.25
        for dim in other_dims:
            # Repeat tiled gaussian structure for other dimensions.
            tiled_means = tiled_means[...,np.newaxis].repeat(dim, axis=-1)

        mean_prior = Normal(tiled_means, np.full_like(tiled_means, 0.2))
        sd_prior = HalfNormal(np.full_like(tiled_means, 0.4))
            
        parameters = Phi(
            Parameter(name='mean', shape=shape, bounds=mean_bounds,
                        prior=mean_prior),
            Parameter(name='sd', shape=shape, bounds=sd_bounds,
                        prior=sd_prior, zero_to_epsilon=True)
            )

        return parameters

    @property
    def coefficients(self):
        """Return N discrete gaussians with T bins, where `shape=(T,N)`."""
        mean = self.parameters['mean'].values
        sd = self.parameters['sd'].values
        n_input_channels = self.shape[0]

        x = np.arange(n_input_channels)/n_input_channels
        mean = np.asanyarray(mean)[..., np.newaxis]
        sd = np.asanyarray(sd)[..., np.newaxis]
        coefficients = np.exp(-0.5*((x-mean)/sd)**2)  # (rank, ..., outputs, T)
        reordered = np.moveaxis(coefficients, -1, 0)  # (T, rank, ..., outputs)
        # Normalize by the cumulative sum for each channel
        cumulative_sum = np.sum(reordered, axis=-1, keepdims=True)
        # If all coefficients for a channel are 0, skip normalization
        cumulative_sum[cumulative_sum == 0] = 1
        normalized = reordered/cumulative_sum

        return normalized

    def evaluate(self, input):
        """Multiply `input[...,i]` by `WeightChannels.coefficients[...,i]`."""

        try:
            output = np.moveaxis(
                np.matmul(np.rollaxis(input, 2), np.rollaxis(self.coefficients, 2)),
                [0,1,2],[2,0,1])
        except ValueError as e:
            # Check for dimension swap, to give more informative error message.
            if 'mismatch in core dimension' in str(e):
                raise ShapeError(self, input=input.shape,
                                 coefficients=self.coefficients.shape)
            else:
                # Otherwise let the original error through.
                raise e

        return output

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        class WeightChannelsMultiTF(NemsKerasLayer):
            @tf.function
            def call(self, inputs):
                # reshape inputs and coefficients so that mult can happen on last
                # two dimensions. Broadcasting seems to work fine this way
                out = tf.transpose(
                    tf.matmul(
                        tf.transpose(inputs, perm=[0, 3, 1, 2]),
                        tf.transpose(self.coefficients, perm=[2, 0, 1])
                        ),
                    perm=[0, 2, 3, 1]
                    )
                return out

        return WeightChannelsMultiTF(self, **kwargs)




from fractions import Fraction

import matplotlib.pyplot as plt
import matplotlib.ticker as mticker


def ax_remove_box(axes=None):
    """Remove right and top lines from plot border on Matplotlib axes."""
    if axes is None:
        axes = plt.gca()
    axes.spines['right'].set_visible(False)
    axes.spines['top'].set_visible(False)

def ax_bins_to_seconds(axes=None, sampling_rate=1.0, time_axis='x',
                       conversion_factor=1.0, decimals=2):
    """Change tick labels on time axis from bins to seconds.
    
    Parameters
    ----------
    axes : matplotlib.axes.Axes; optional.
        Axes to modify. If not provided, `plt.gca()` will be used.
    sampling_rate : float; default=1.0.
        Rate (in Hz) at which data on axes were collected.
    time_axis : str; default='x'.
        'x' or 'y', the axis on `axes` that represents time.
    conversion_factor : int or float; default=1.0.
        Multiply seconds by this number to get different units. Floating point
        values must correspond to rational numbers. Scientific notation may only
        be used for multiples of 10, and will be converted to an integer.
        Ex: `conversion_factor=1000` to get units of milliseconds.
            `conversion_factor=1/60` to get units of minutes.
    decimals : int; default=2.
        Number of decimal places to show on new tick labels.
    
    """
    if axes is None:
        axes = plt.gca()
    if (conversion_factor % 10) == 0:
        # re-cast scientific notation
        conversion_factor = int(conversion_factor)  

    # Get axes methods based on time_axis (x or y)
    tick_getter = getattr(axes, f'get_{time_axis}ticks')
    axis = getattr(axes, f'{time_axis}axis')
    set_tick_labels = getattr(axes, f'set_{time_axis}ticklabels')
    set_label = getattr(axes, f'set_{time_axis}label')

    # Fetch list of current tick locations (which have units of bins).
    # Set locations to be fixed, and update the labels at each location
    ticks = tick_getter().tolist()
    axes.xaxis.set_major_locator(mticker.FixedLocator(ticks))
    labels = [f'{(tick/sampling_rate)*conversion_factor:.{decimals}f}'
               for tick in ticks]
    set_tick_labels(labels)

    if conversion_factor == 1:
        units = ''  # still seconds
    else:
        # Format 1/conversion_factor as a fraction to represent units.
        if conversion_factor < 1:
            denom = 1/conversion_factor
        else:
            denom = conversion_factor
        units = f'{Fraction(1/conversion_factor).limit_denominator(denom)} '
    set_label(f'Time ({units}s)')


def standardize_axes(axes, sampling_rate=None, time_kwargs=None, x_margin=False):
    """Apply NEMS formatting to a matplotlib Axes in-place.

    Dependent upon specified parameters, this function will:
        - Remove the right and top axes border.
        - Remove extra whitespace on the x-axis.
        - Convert units of bins to seconds, based on `sampling_rate`.

    Parameters
    ----------
    sampling_rate : float; optional.
        If specified, time axis will be labeled with units of seconds.
    time_kwargs : dict; optional.
        Additional keyword arguments for `ax_bins_to_seconds`.
    x_margin : bool; default=False.
        If False, remove whitespace margins from xlim
        (i.e. timeseries will fill the full x-axis)
    
    """
    if time_kwargs is None: time_kwargs = {}
    if sampling_rate is not None:
        # convert bins to seconds
        ax_bins_to_seconds(axes, sampling_rate, **time_kwargs)
    ax_remove_box(axes)    # remove right and top box borders
    if not x_margin:
        axes.set_xmargin(0)  # remove white space around xlim



class DataMap:
    def __init__(self, layer):
        self.layer = layer
        self.name = layer.name
        self.input = layer.input
        if not isinstance(self.input, list):
            self.input = [self.input]
        self.output = layer.output

        # Special check for adding state kwarg when no input is specified
        self.state_arg = layer.state_arg
        self.add_state = False
        if (layer.input is None) and (layer.state_arg is not None):
            self.add_state = True

        self.state = layer.state_arg
        self.out = []
        self._build_input_map()  # assigns self.args, self.kwargs

    def _build_input_map(self):
        """Map `Layer.input` to *args and **kwargs for `Layer.evaluate`.
        
        Adds keys to `self.args` and `self.kwargs`, corresponding to keys in
        `data`.
        
        """
        args = []
        kwargs = {}
        for key in self.input:
            if isinstance(key, dict):
                kwargs.update({k: v for k, v in key.items()})
            else:
                args.append(key)
        if self.add_state:
            # This will be substituted with the actual key when fetching arrays
            kwargs[self.state_arg] = '_state_name'

        self.args = args
        self.kwargs = kwargs

    def map_outputs(self, eval_out, output_name=None):
        """Determine where/whether to save `eval_out` in `data`.

        Adds keys to `self.out`. Entries will be all None (`eval_out` is not
        saved) or all strings (`data[string] = eval_out`). If `self.output` is
        a string and `len(eval_out) > 1`, then that string will be incremented
        in the form of `string.{i}` to get a unique key for each output.
        
        """
        if not isinstance(eval_out, (list, tuple)):
            eval_out = [eval_out]
        output = self.output
        if (output_name is not None) and (output is None):
            # Intended for use by `Model.evaluate` on final layer.
            output = output_name

        if output is not None:
            if isinstance(output, str):
                output_keys = [
                    f"{self.output}.{i}" if i != 0 else f"{output}"
                    for i in range(len(eval_out))
                    ]
            elif isinstance(output, list):
                if len(output) != len(eval_out):
                    raise ValueError(
                        f"Layer {self.layer.name} specified "
                        f"{len(output)} outputs, but .evaluate returned "
                        f"{len(eval_out)} outputs."
                    )
                output_keys = output
            else:
                raise TypeError(
                    f"Unrecognized type for {self.name}.output:"
                    f"{type(output)}"
                    )
        else:
            output_keys = [None]*len(eval_out)
        self.out = output_keys

    def get_inputs(self, data):
        """Get arrays from `data` that correspond to inputs for `Layer.evaluate`.
        
        Parameters
        ----------
        data : dict of ndarray.
            Input data for Model. See `Model.evaluate` for details on structure.
        input_name : str
            Key for input to first layer. Only used if `Layer.input is None`
            and `data['last_output']` is None.

        Returns
        -------
        (args, kwargs, out) : (list, dict, list)
            Entries are ndarrays pulled from `data`.

        Warnings
        --------
        The output of this method is dependent upon the state of `data`, and
        is designed to be used as part of the Model.evaluate` loop. Otherwise,
        care must be taken to ensure that the correct keys are present in `data`
        and are associated with the correct arrays.
        
        """
        last_out = data.get('_last_output', None)
        if (last_out is None) and (self.layer.input is None):
            # First statement should only be True for first Layer
            self.args = [data.input_name]
        missing_input = ("Cannot determine Layer input in Model.evaluate, \n"
                         f"specify `last_out` or `{self.name}.input`.")

        args = []
        for a in self.args:
            if a is None:
                # Supply each of the previous outputs as a separate input,
                # i.e. `layer.evaluate(*last_out)`.
                if last_out is None:
                    raise ValueError(missing_input)
                args.append(last_out)
            elif isinstance(a, list):
                # Supply a list of arrays at all keys as a single argument,
                # i.e. `layer.evaluate([array1, array2, ...])`.
                args.append([data[k] for k in a])
            else: # (a should be a string)
                # Add one array to arguments,
                # i.e. `layer.evaluate(array)`.
                args.append(data[a])

        update_state=None
        kwargs = {}
        for k, v in self.kwargs.items():
            if v is None:
                # Supply keyword `k` with a list of the previous outputs,
                # i.e. `layer.evaluate(k=[out1, out2, ...])`.
                if last_out is None:
                    raise ValueError(missing_input)
                kwargs[k] = last_out
            elif isinstance(v, list):
                # Supply keyword `k` with a list of the indicated arrays,
                # i.e. `layer.evaluate(k=[array1, array2, ...])`.
                kwargs[k] = [data[_k] for _k in v]
            else: # (v should be a string)
                # Supply keyword `k` with a single array,
                # i.e. `layer.evaluate(k=array1)`.
                if v == '_state_name':
                    # Special check for inserting state array
                    v = data.state_name  # replace _state_name with actual key
                    update_state = {k: v}
                kwargs[k] = data[v]
        
        # Replace _state_name
        if update_state is not None:
            self.kwargs.update(update_state)

        return args, kwargs

    @property
    def keys(self):
        return {'args': self.args, 'kwargs': self.kwargs, 'out': self.out}

    def __repr__(self):
        name = self.layer.name
        if name is None:
            name = self.layer.default_name
            
        s = f"DataMap(layer={name})\n"
        s += f".args: {self.args}\n"
        s += f".kwargs: {self.kwargs}\n"
        s += f".out: {self.out}\n"
        s += f"*"*16 + "\n"
        return s

    def __iter__(self):
        """Return an iter-wrapped singleton list containing this DataMap.

        Supports compatibility for iterating over DataMaps returned by
        `Model.get_data_maps` when single Layers are indexed, without the need
        to repeatedly add `[0]` to the end of indexing statements.
        
        NOTE: This does *not* iterate over the DataMap itself.
        TODO: This feels kludgy and I don't like it, but it allows for some
              nice syntax for navigating Models.
        
        """
        return iter([self])



"""Miscellaneous utilities for manipulating NumPy ndarrays."""

import numpy as np
from numba import njit


# Based on solution by:
# user48956
# https://stackoverflow.com/questions/911871/detect-if-a-numpy-array-contains-at-least-one-non-numeric-value
@njit(nogil=True)
def one_or_more_nan(array):
    """True if an array contains at least one NaN value.
    
    Much faster than native numpy solutions if a NaN occurs 'early', and
    not much slower if no NaNs are present.
    
    """
    for value in array.flat:
        if np.isnan(value): return True
    else:
        return False


@njit(nogil=True)
def one_or_more_negative(array):
    """True if an array contains at least one negative number."""
    for value in array.flat:
        if value < 0: return True
    else:
        return False


def broadcast_axis_shape(array1, array2, axis=0):
    """Get shape where `array1.shape[axis] == array2.shape[axis]`.
    
    Parameters
    ----------
    array1 : np.ndarray
        Array to be broadcast.
    array2 : np.ndarray
        Array to broadcast to.
    axis : int; default=0.
        Axis of array to broadcast.
    
    Returns
    -------
    shape : tuple
        Broadcasted shape for `array1`.

    """
    shape = list(array1.shape)
    shape[axis] = array2.shape[axis]
    return tuple(shape)


def broadcast_axes(array1, array2, axis=0):
    """Broadcast `array1` to `array2.shape`, but only along specified `axis`.
    
    Parameters
    ----------
    array1 : np.ndarray
        Array to be broadcast.
    array2 : np.ndarray
        Array to broadcast to.
    axis : int; default=0.
        Axis of array to broadcast.

    Returns
    -------
    new_array : np.ndarray
        Broadcasted version of `array1`.

    """
    broadcasted_shape = broadcast_axis_shape(array1, array2, axis=axis)
    if broadcasted_shape[axis] == array1.shape[axis]:
        # Don't waste time broadcasting if the shape is the same.
        new_array = array1
    else:
        new_array = np.broadcast_to(array1, broadcasted_shape)

    return new_array


def broadcast_dicts(d1, d2, axis=0, debug_memory=False):
    """Broadcast axis length of all arrays in one dict against another dict.

    Parameters
    ----------
    d1 : dict of np.ndarray.
        Dictionary containing arrays to be broadcast.
    d2 : dict of np.ndarray.
        Dictionary containing arrays to broadcast to.
    axis : int; default=0.
        Axis to broadcast.
    debug_memory : bool; default=False.
        If True, raise AssertionError if broadcasted arrays do not share memory
        with originals.

    Returns
    -------
    dict of np.ndarray.
    
    """

    new_d = {}
    for k, v in d1.items():
        if v.shape[axis] != 1:
            # Can't broadcast
            new_d[k] = v
        else:
            for v2 in d2.values():
                if v2.shape[axis] > 1:
                    # Broadcasting is possible.
                    new_v = broadcast_axes(v, v2, axis=axis)
                    if debug_memory:
                        assert np.shares_memory(new_v, v)
                    new_d[k] = new_v
                    # Shape on axis no longer 1, can only broadcast once.
                    break
            else:
                if k not in new_d:
                    # No compatible arrays for broadcasting.
                    new_d[k] = v
    
    if len(new_d) == 0:
        # There was nothing to broadcast
        new_d = d1.copy()

    return new_d


def concatenate_dicts(*dicts, axis=0, newaxis=False):
    """Concatenate key-matched arrays in `dicts` along `axis`.
    
    Parameters
    ----------
    dicts : N-tuple of dict of np.ndarray.
        Each dict should have the format `{k: arrary}`. All dicts must contain
        the same keys. All arrays at the same key must have the same shape on
        all dimensions except `axis`.
    newaxis : bool; default=False.
        If True, use `np.stack` instead of `np.concatenate`, which will add
        a new axis to concatenate on at the indicated position. All dimensions
        of arrays at the same key must have the same shape, without exception.

    Returns
    -------
    dict of ndarray, in the format of `{k: concatenated_array}`.

    Notes
    -----
    This function will output array copies, even if the arrays in `dicts` are
    views. So far as I can tell, this is an unavoidable limitation from NumPy
    due to the fact that views can be non-continguous in memory, while the
    concatenated arrays must be contiguous.
    
    """

    listed = {k: [d[k] for d in dicts] for k in dicts[0].keys()}
    if newaxis:
        fn = np.stack
    else:
        fn = np.concatenate
    concatenated = {k: fn(v, axis=axis) for k, v in listed.items()}

    return concatenated


def apply_to_dict(fn, d, *args, allow_copies=True, **kwargs):
        """Maps {k: v} -> {k: fn(v, *args, **kwargs)} for all k, v in dict `d`.
        
        Parameters
        ----------
        fn : callable
            Must accept a single ndarray as its first positional argument.
        d : dict of ndarrays.
            Dictionary containing arrays that `fn` will be applied to.
        args : N-tuple
            Additional positional arguments for `fn`.
        allow_copies : bool; default=True.
            If False, raise AssertionError if `fn(v, *args, **kwargs)` returns
            an array that does not share memory with `v`. Useful for debugging
            if you're uncertain whether `fn` will return copies or views.
        kwargs : dict
            Additional keyword arguments for `fn`.

        Returns
        -------
        dict of ndarray
            Same keys as `d`, with values replaced by output of `fn`.

        Examples
        --------
        TODO
        
        """
        new_d = {}
        for k, v in d.items():
            new_v = fn(v, *args, **kwargs)
            if not allow_copies:
                assert np.shares_memory(new_v, v)
            new_d[k] = new_v
        return new_d



"""Implements DataSet container for use by Model.evaluate.

TODO: May be a better place to put this?

"""

import numpy as np

from nems.tools.arrays import broadcast_dicts, concatenate_dicts, apply_to_dict


class DataSet:
    # Use these names if `input_name`, `state_name`, etc. are not specified
    # in DataSet.__init__.
    default_input = 'input'
    default_output = 'output'
    default_target = 'target'
    default_state = 'state'

    def __init__(self, input, state=None, target=None, input_name=None,
                 state_name=None, output_name=None, target_name=None,
                 dtype=None, has_samples=False, debug_memory=False,
                 **kwargs):
        """Container for tracking dictionaries of data arrays.
        
        See `Model.evaluate` and `Model.fit` for detailed documentation of
        parameters.

        Parameters
        ----------
        input : ndarray or dict.
        state : ndarray; optional.
        target : ndarray or dict; optional.
        input_name : str; optional.
        state_name : str; optional.
        output_name : str; optional.
        target_name : str; optional.
        dtype : type; optional.
            TODO: WIP. Want to specify a consistent datatype to cast all
                  arrays to.
        has_samples : bool; default=False.
            Indicates if first dimension of inputs (and other optional data)
            represents samples, saved as `DataSet.has_samples`. Methods that
            add or assume a sample dimension will set this attribute to True,
            and methods that remove a sample dimension will set this to False.
            In most cases, this kwarg should not need to be specified at time
            of construction.
        debug_memory : bool; default=False.
            TODO: Not actually implemented yet.
            If True, check `np.shares_memory(array, copied_array)` in several
            places to ensure that memory is not duplicated unintentionally.
        kwargs : dict; optional.
            Extra kwargs are silently ignored for convenience, so that other
            code can use `DataSet(input, **evaluate_kwargs)`.
        
        Attributes
        ----------
        inputs : dict.
            All model inputs, stored as `{key: ndarray}`.
        outputs : dict.
            All saved Layer outputs, stored in the same format. Special key
            '_last_output' will contain the most recent Layer output during
            evaluation, but this key is removed by `DataSet.finalize_data`.
            Will be empty if `save_output` has never been called.
        targets : dict.
            All optimization targets, stored in the same format. Will be empty
            if `target is None`.

        """
        # Set self.<attr>_name to default if <attr>_name is None, otherwise
        # save self.<attr>_name.
        names = zip(['input', 'state', 'output', 'target'],
                    [input_name, state_name, output_name, target_name])
        for attr, name in names:
            if name is None: name = getattr(self, f'default_{attr}')
            setattr(self, f'{attr}_name', name)

        self.has_samples = has_samples
        self.debug_memory = debug_memory
        self.initialize_data(input, state, target)  # other kwargs too

        if dtype is None:
            # Set to same as first input.
            self.dtype = list(self.inputs.values())[0].dtype
        else:
            self.dtype = dtype

    @property
    def n_samples(self):
        """Get size of first dimension of first input.
        
        If `DataSet.has_samples` is `False`, this method returns `None` to
        indicate that the number of samples is not known. Either there is no
        sample dimension, or there is but `DataSet` doesn't know about it.
        
        """
        if self.has_samples:
            n_samples = list(self.inputs.values())[0].shape[0]
        else:
            n_samples = None
        return n_samples

    def initialize_data(self, input, state=None, target=None):
        """Package data into dictionaries, store in attributes.
        
        Assigns data to `DataSet.inputs, DataSet.outputs, DataSet.targets`.

        Parameters
        ----------
        input : ndarray or dict.
        state : ndarray; optional.
        target : ndarray or dict; optional.
        
        """

        # Initialize inputs
        if isinstance(input, np.ndarray):
            input_dict = {self.input_name: input}
            if state is not None:
                input_dict[self.state_name] = state
        else:
            # Arrays in shallow copy will share memory, but the new data
            # dictionary will end up with additional keys after evaluation.
            input_dict = input.copy()

        # Initialize outputs
        output_dict = {}

        # Initialize targets
        if target is None:
            target_dict = {}
        elif isinstance(target, np.ndarray):
            target_dict = {self.target_name: target}
        else:
            target_dict = target.copy()

        # Make sure all targets are at least 2D, otherwise undesired broadcasting
        # may occurr in cost functions.
        target_dict = apply_to_dict(
            lambda a: a if a.ndim >= 2 else a[..., np.newaxis],
            target_dict, allow_copies=False
            )

        self.inputs = input_dict
        self.outputs = output_dict
        self.targets = target_dict

    def save_output(self, keys, output):
        """Save `output` in `DataSet.outputs` for keys that are not `None`.
        
        Parameters
        ----------
        keys : list of str.
            Indicates keys in `DataSet.outputs` where array(s) in `output`
            should be stored. If a key is None, the output is not saved
            (except at special key `_last_output`).
        output : ndarray or list of ndarray.
            Output of a call to Layer._evaluate.
        
        """
        # keys is always a list, but output might be a list or one array.
        if isinstance(output, (list, tuple)):
            self.outputs.update({k: v for k, v in zip(keys, output)
                        if k is not None})
        elif keys[0] is not None:
            self.outputs[keys[0]] = output
        # Always save output to _last_output for use by Model.evaluate
        self.outputs['_last_output'] = output

    def finalize_data(self, final_layer):
        """Remove special keys from `DataSet.outputs` and update final_layer.
        
        If `final_layer.output is None`, save the Layer's output to
        `DataSet.output_name` instead and update `Layer.data_map` to reflect
        the new key(s).
        
        """
        # Re-name last output if keys not specified by Layer
        final_output = self.outputs['_last_output']
        if final_layer.output is None:
            # Re-map `data_map.out` so that it reflects `output_name`.
            final_layer.data_map.map_outputs(final_output, self.output_name)
            self.save_output(final_layer.data_map.out, final_output)
        _ = self.outputs.pop('_last_output')

    def as_broadcasted_samples(self):
        """Broadcasts all data arrays against each other for number of samples.
        
        Arrays with shape (1, T, ..., N), where T is the number of time bins
        and N is the number of output channels, will broadcast to shape
        `(S, T, ..., N)`, where S is the number of samples in other data arrays.
        First all inputs are broadcast against other inputs (and outputs against
        other outputs, etc) and then inputs are broadcast against outputs and
        targets, outputs against inputs and targets, etc.

        Examples
        --------
        >>> stimulus = np.random.rand(1, 1000, 18)  # one stimulus sample
        >>> response = np.random.rand(10, 1000, 18) # multiple trials
        >>> data = DataSet(input=stimulus, target=response)
        >>> broadcasted = data.as_broadcasted_samples()
        >>> broadcasted.inputs['stimulus'].shape
        (10, 1000, 18)
        >>> np.shares_memory(broadcasted.inputs['stimulus'], stimulus)
        True
        
        """

        # In case inputs/outputs and targets have different numbers of samples,
        # broadcast within each category first.
        inputs, outputs, targets = [
            broadcast_dicts(d1, d1, debug_memory=self.debug_memory)
            for d1 in [self.inputs, self.outputs, self.targets]
        ]

        # Then broadcast each category to the others.
        inputs, outputs, targets = [
            broadcast_dicts(d1, d2, debug_memory=self.debug_memory)
            for d1, d2 in [
                (inputs, {**outputs, **targets}),
                (outputs, {**inputs, **targets}),
                (targets, {**inputs, **outputs})
                ]
        ]

        copy = self.modified_copy(inputs, outputs, targets)
        copy.has_samples = True

        return copy

    def as_batches(self, batch_size=None, permute=False):
        """Generate copies of DataSet containing single batches.
        
        Parameters
        ----------
        batch_size : int; optional.
            Number of samples to include in each batch. If the total number of
            samples is not evenly divisble by `batch_size`, then one batch will
            have fewer samples. If `batch_size` is None, `batch_size` will be
            set equal to the total number of samples.
        permute : bool; default=False.
            TODO: Not implemented yet.
            If True, shuffle yield order of batches.

        Yields
        ------
        DataSet
            Dict entries will be lists of samples.

        Notes
        -----
        This implementation results in a list of views into the original data
        (i.e. memory is shared). If changes are made, make sure the new version
        doesn't result in copies (which could increase memory usage
        dramatically).

        Warnings
        --------
        This method assumes the first dimension of all data arrays represents
        samples. Unexpected behavior will result if this assumption is invalid.
        
        """

        # TODO: This handles cases of 1 sample -> many samples or vise-versa,
        #       but do we want to support n samples -> m samples? Would have
        #       to do some kind of tiling that may have side-effects that I'm
        #       not thinking of, and I'm not sure if this would be useful.
        d = self.as_broadcasted_samples()
        d.has_samples = True
        
        # Split data into batches along first axis. Should end up with a list
        # of arrays with shape (B, T, N), where B is `batch_size` (i.e. number
        # of samples per batch), stored at each key.
        batched_inputs, batched_outputs, batched_targets = [
            d._arrays_to_batches(_dict, batch_size)
            for _dict in [d.inputs, d.outputs, d.targets]
        ]

        n_batches = len(list(batched_inputs.values())[0])

        # Index into batched_data instead of collecting a list of batches,
        # to ensure memory is still shared. Also makes permutations easier.
        indices = np.arange(n_batches)
        if permute:
            # Randomly shuffle indices
            np.random.shuffle(indices)
            # TODO: Not quite this simple, have to be able to put the concatenated
            #       outputs back in the right order. So need to store the shuffled
            #       indices somehow.
            # TODO: Should samples be shuffled instead/in addition? Otherwise
            #       the same samples always end up in a batch together.
            raise NotImplementedError("Shuffling batches not implemented yet")

        for i in indices:
            inputs = {k: v[i] for k, v in batched_inputs.items()}
            outputs = {k: v[i] for k, v in batched_outputs.items()}
            targets = {k: v[i] for k, v in batched_targets.items()}
            d.assert_no_copies(inputs, outputs, targets)
            yield d.modified_copy(inputs, outputs, targets)

    def _arrays_to_batches(self, data, batch_size):
        """Internal for `as_batches`.
        
        Parameters
        ----------
        data : dict
        batch_size : int

        Returns
        -------
        dict
        
        """

        if (batch_size is None) and (len(data) > 0):
            # Assume sample dimension exists, set batch_size to force 1 batch
            batch_size = list(data.values())[0].shape[0]
        batched_data = {
            k: np.split(v, np.arange(batch_size, len(v), batch_size))
            for k, v in data.items()
            }

        return batched_data

    def as_samples(self):
        """Generate copies of a batch DataSet, containing single samples.
        
        `DataSet.inputs, .outputs, .targets` should contain a list of arrays at
        each key, as yielded by `DataSet.as_batches`.

        Yields
        ------
        DataSet

        Warnings
        --------
        If this method is used on a DataSet that is *not* in the format yielded
        by `DataSet.as_batches`, unexpected behavior will result.

        """

        n_samples = len(list(self.inputs.values())[0])
        s_inputs, s_outputs, s_targets = [
            {k: np.split(v, n_samples) for k, v in d.items()}
            for d in [self.inputs, self.outputs, self.targets]
            ]

        for i in range(n_samples):
            inputs = {k: v[i].squeeze(axis=0) for k, v in s_inputs.items()}
            outputs = {k: v[i].squeeze(axis=0) for k, v in s_outputs.items()}
            targets = {k: v[i].squeeze(axis=0) for k, v in s_targets.items()}
            if self.debug_memory:
                self.assert_no_copies(inputs, outputs, targets)
            sample = self.modified_copy(inputs, outputs, targets)
            sample.has_samples = False
            yield sample

    def prepend_samples(self):
        """Prepend a singleton sample dimension."""
        data = self.apply(lambda v: v[np.newaxis,...], allow_copies=False)
        data.has_samples = True
        return data

    def squeeze_samples(self):
        """Remove singleton sample dimension from all arrays."""
        data = self.apply(lambda v: np.squeeze(v, axis=0), allow_copies=False)
        data.has_samples = False
        return data

    @staticmethod
    def concatenate_sample_outputs(data_sets):
        """Concatenate key-matched arrays from each `data_set.outputs`.
        
        Parameters
        ----------
        data_sets : list of DataSet

        Returns
        -------
        dict of ndarray, in the format of `DataSet.outputs`.
        
        """
        return concatenate_dicts(*[d.outputs for d in data_sets])

    def modified_copy(self, inputs, outputs, targets):
        """Get a shallow copy of DataSet with new data dictionaries.
        
        Parameters
        ----------
        inputs : dict
        outputs : dict
        targets : dict

        Returns
        -------
        DataSet
        
        """
        data = DataSet(
            inputs, state=None, target=targets, input_name=self.input_name,
            state_name=self.state_name, output_name=self.output_name,
            target_name=self.target_name, dtype=self.dtype,
            has_samples=self.has_samples,
            )
        data.outputs = outputs
        return data

    def copy(self):
        """Get a shallow copy of DataSet."""
        return self.modified_copy(self.inputs, self.outputs, self.targets)

    def apply(self, fn, *args, allow_copies=False, inplace=False, **kwargs):
        """Maps {k: v} -> {k: fn(v, *args, **kwargs)} for all k, v in DataSet.
        
        Parameters
        ----------
        fn : callable
            Must accept a single ndarray as its first positional argument.
        args : N-tuple
            Additional positional arguments for `fn`.
        allow_copies : bool; default=True.
            If False, raise AssertionError if `fn(v, *args, **kwargs)` returns
            an array that does not share memory with `v`.
        kwargs : dict
            Additional keyword arguments for `fn`.

        Returns
        -------
        DataSet
            A modified copy containing the transformed arrays.

        Examples
        --------
        TODO
        
        """
        inputs, outputs, targets = [
            apply_to_dict(fn, d, *args, allow_copies=allow_copies, **kwargs)
            for d in [self.inputs, self.outputs, self.targets]
            ]
        if inplace:
            self.inputs = inputs
            self.outputs = outputs
            self.targets = targets
        else:
            return self.modified_copy(inputs, outputs, targets)

    def assert_no_copies(self, inputs, outputs, targets):
        """Check if arrays in dictionaries share memory with arrays in DataSet.
        
        Useful for debugging memory inflation. Note that keys of arguments are
        iterated over, not keys of DataSet. This means inputs, outputs and
        targets can contain subsets of the keys present in DataSet, in which
        case not all arrays in DataSet will be checked.

        Parameters
        ----------
        inputs : dict
        outputs : dict
        targets : dict

        Raises
        ------
        AssertionError
            If `np.shares_memory(inputs[k], DataSet.inputs[k])` is False for
            any array in inputs. The same comparison is repeated for outputs
            and targets.

        """
        for k in inputs.keys():
            assert np.shares_memory(inputs[k], self.inputs[k])
        for k in outputs.keys():
            assert np.shares_memory(outputs[k], self.outputs[k])
        for k in targets.keys():
            assert np.shares_memory(targets[k], self.targets[k])

    def as_dict(self):
        """Get `DataSet.inputs, .outputs, .targets` as a single dictionary."""
        return {**self.inputs, **self.outputs, **self.targets}

    def as_type(self, dtype, inplace=False):
        """Replace all arrays with a different dtype."""
        # NOTE: Even with `copy=False`, this will result in a copy if dtype
        #       changes precision (e.g. float64 -> float32). 
        return self.apply(lambda a: a.astype(dtype, copy=False), inplace=inplace,
                          allow_copies=True)

    # Pass dict get (but not set) operations to self.inputs, outputs, targets
    def __getitem__(self, key):
        return self.as_dict()[key]
    def get(self, key, default):
        return self.as_dict().get(key, default)
    def items(self):
        return self.as_dict().items()
    def __iter__(self):
        return self.as_dict().__iter__()
    def __len__(self):
        return len(self.as_dict())



# This code is derived from the gammatone toolkit, licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING

"""Renders spectrograms which use gammatone filterbanks."""

from __future__ import division
import numpy as np

from .filters import make_erb_filters, centre_freqs, erb_filterbank


def gammagram(wave, fs=44000, window_time=0.01, hop_time=0.01, channels=18,
              f_min=200.0, f_max=None):
    """Calculate a spectrogram-like array based on gammatone subband filters.
    
    The waveform `wave` (at sample rate `fs`) is passed through a multi-channel
    gammatone auditory model filterbank, with lowest frequency `f_min` and
    highest frequency `f_max`. The outputs of each band then have their energy
    integrated over windows of `window_time` seconds, advancing by `hop_time`
    seconds for successive columns. These magnitudes are returned as a
    nonnegative real matrix with `channels` rows.

    Parameters
    ----------
    wave : np.ndarray.
        Sound waveform with shape (T,) or (T,1).
    fs : int; default=44000.
        Sampling frequency. `gtgram_strides` uses this to scale `window_time`
        and `hop_time` to convert them to units of bins.
    window_time : float; default=0.01.
        Length of integration window. Default of 0.01 corresponods to 1/100hz,
        where 100hz is the most common spike raster sampling rate used by LBHB.
    hop_time : float; default=0.01.
        Stepsize of window advancement for successive columns. Default of 0.01
        corresponods to 1/100hz, where 100hz is the most common spike raster
        sampling rate used by LBHB.
    channels : int.
        Number of frequency channels in the spectrogram-like output.
    f_min : float; default=200.0.
        Lower frequency cutoff.
    f_max : float; optional.
        Upper frequency cutoff.

    Returns
    -------
    np.ndarray
        With shape (T, `channels`)
    
    Copyright
    ---------
    2009-02-23 Dan Ellis dpwe@ee.columbia.edu
    (c) 2013 Jason Heeris (Python implementation)

    """

    xe = _gtgram_xe(wave, fs, channels, f_min, f_max)
    
    nwin, hop_samples, ncols = gtgram_strides(
        fs,
        window_time,
        hop_time,
        xe.shape[1]
    )
    
    y = np.zeros((channels, ncols))
    
    for cnum in range(ncols):
        segment = xe[:, cnum * hop_samples + np.arange(nwin)]
        y[:, cnum] = np.sqrt(segment.mean(1))
    
    return y


def _gtgram_xe(wave, fs, channels, f_min, f_max=None, verbose=False):
    """Calculate the intermediate ERB filterbank processed matrix.
    
    Internal for `gammagram`.
    
    """

    cfs = centre_freqs(fs, channels, f_min, f_max)
    if verbose:
        print('cfs: ', cfs)
    fcoefs = np.flipud(make_erb_filters(fs, cfs))
    xf = erb_filterbank(wave, fcoefs)
    xe = np.power(xf, 2)

    return xe


def gtgram_strides(fs, window_time, hop_time, filterbank_cols):
    """Calculate the window size for a gammatone filter spectrogram.
    
    Parameters
    ----------
    fs : int.
        Sampling frequency.
    window_time : float; default=0.01.
        Length of integration window. Default of 0.01 corresponods to 1/100hz,
        where 100hz is the most common spike raster sampling rate used by LBHB.
    hop_time : float; default=0.01.
        Stepsize of window advancement for successive columns. Default of 0.01
        corresponods to 1/100hz, where 100hz is the most common spike raster
        sampling rate used by LBHB.
    filterbank_cols : int.

    Returns
    -------
    (window_size, hop_samples, output_columns)

    """

    nwin        = int(round_half_away_from_zero(window_time * fs))
    hop_samples = int(round_half_away_from_zero(hop_time * fs))
    columns     = int(np.floor((filterbank_cols - nwin)/hop_samples)) + 1

    return (nwin, hop_samples, columns)


def round_half_away_from_zero(num):
    """Implements the "round-half-away-from-zero" rule.
    
    Fractional parts of 0.5 result in rounding up to the nearest positive
    integer for positive numbers, and down to the nearest negative number for
    negative integers.

    Parameters
    ----------
    num : float

    Returns
    -------
    int

    """
    return np.sign(num) * np.floor(np.abs(num) + 0.5)



from scipy import stats
import numpy as np

from .base import Distribution


class Beta(Distribution):
    """Beta prior.

    Parameters
    ----------
    alpha : scalar or ndarray
        Parameter of distribution
    beta : scalar or ndarray
        Parameter of distribution

    """

    def __init__(self, alpha, beta):
        self._alpha = np.asarray(alpha)
        self._beta = np.asarray(beta)

        # The beauty of using the scipy stats module is it correctly handles
        # scalar and multidimensional arrays of distributions.
        self.distribution = stats.beta(self._alpha, self._beta)

    def __repr__(self):
        alpha = self.value_to_string(self._alpha)
        beta = self.value_to_string(self._beta)
        return 'Beta(α={}, β={})'.format(alpha, beta)
        


from scipy import stats
import numpy as np

from .base import Distribution


class Gamma(Distribution):
    """Gamma prior.

    Parameters
    ----------
    alpha : scalar or ndarray
        Parameter of distribution
    beta : scalar or ndarray
        Parameter of distribution

    Examples
    --------
    To create a Gamma prior using the mean and standard deviation, use the
    classmethod, `from_moments`. This will solve for `alpha` and `beta` given
    your desired mean and standard deviation.
    >>> prior = Gamma.from_moments(mu=10, sd=100)

    """

    @classmethod
    def from_moments(cls, mu, sd):
        # E[X] = α/β
        # Var[X] = α/β^2
        var = sd**2
        alpha = (mu**2)/var
        beta = alpha/mu
        return cls(alpha, beta)

    def __init__(self, alpha, beta):
        self._alpha = np.asarray(alpha)
        self._beta = np.asarray(beta)
        self.distribution = stats.gamma(a=self._alpha, scale=1/self._beta)

    def __repr__(self):
        alpha = self.value_to_string(self._alpha)
        beta = self.value_to_string(self._beta)
        return 'Gamma(α={}, β={})'.format(alpha, beta)



import ast

import numpy as np

from nems.registry import layer
from .base import Layer

class NumPy(Layer):
    """Applies a NumPy operation to inputs.
    
    Parameters
    ----------
    operation : str
        Name of `np.{operation}` to apply to data, ex: 'concatenate'.
    op_args : N-tuple; optional.
        Positional arguments to supply to `np.{operation}`. These do not
        include arrays supplied through `inputs` in `evaluate`.
    op_kwargs : dict; optional.
        Keyword arguments to supply to `np.{operation}`.
    input, output, name : see docs for `Layer`.
    
    Returns
    -------
    ndarray or list of ndarray (see docs for {operation} at numpy.org).

    Notes
    -----
    Unlike most Layer subclasses, NumPy Layers will not pass additional
    keyword arguments to `Layer.__init__`. Instead, they are passed to
    `np.{concatenate}`. Additionally, note that `parameters`, `bounds`,
    `priors`, and `shape` cannot be set through initialization and have no
    effect since NumPy Layers do not have any fittable parameters.

    Examples
    --------
    >>> x = np.arange(8).reshape(2,4)
    >>> x
    array([[ 0,  1,  2,  3],
            [ 4,  5,  6,  7]])
    >>> split = NumPy('split', 2, axis=0)
    >>> split.evaluate(x)
    [array([[0, 1, 2, 3]]), array([[4, 5, 6, 7]])]

    """

    def __init__(self, operation, *op_args, input=None, output=None, name=None,
                 **op_kwargs):

        if name is None:
            name = f'np.{operation}'
        super().__init__(input=input, output=output, name=name)

        self._operation = getattr(np, operation)
        self._args = op_args
        op_kwargs = {} if op_kwargs is None else op_kwargs
        self._kwargs = op_kwargs

    def evaluate(self, *inputs):
        """Apply `np.{operation}` to inputs."""
        return self._operation(*inputs, *self._args, **self._kwargs)

    @layer('np')
    def from_keyword(keyword):
        """Construct NumPy Layer from keyword.

        Keyword options
        ---------------
        {operation} : Name of numpy operation to use, ex: 'split', 'sum'.
            NOTE: This must be the first option, ex: `np.split.op2.op3...`
        {value} : Positional arguments for `np.{operation}`.
            Only literal, non-alpha value types are supported, such as `int`,
            `tuple`, or `float` (see docs for `ast.literal_eval`). For an alpha
            string argument (ex: 'full', 'same'), instantiate directly from
            the NumPy Layer class.
        {key}{value} : Keyword arguments for `np.{operation}`.
            Only keys that are all alpha characters (like `axis`, `shape`, etc)
            are supported. For kwarg names with underscores, numeric characters,
            etc, instantiate the Layer directly.
        See also
        --------
        Layer.from_keyword
        
        """
        # TODO: want to be able to do 'np.concatenate.axis1'
        options = keyword.split('.')
        operation = options[1]  # NOTE: this is hard-coded

        args = []
        kwargs = {}
        for op in options[2:]:
            key = ''
            while op[0].isalpha():
                key += op[0]
                op = op[1:]
            val = ast.literal_eval(op)
            if key == '':
                # treat as positional argument
                args.append(val)
            else:
                # keyword argument
                kwargs[key] = val

        return NumPy(operation, *args, **kwargs)



import numpy as np

from nems.distributions import Normal


class Parameter:
    """Stores and manages updates to values for one parameter of one Layer."""

    def __init__(self, name, shape=(), prior=None, bounds=None,
                 default_bounds='infinite', zero_to_epsilon=False,
                 initial_value='mean'):
        """Stores and manages updates to values for one parameter of one Layer.

        Parameters are intended to exist as components of a parent Phi instance,
        by invoking `Phi.add_parameter`. Without establishing this relationship,
        most Parameter methods will not work.

        As with Phi, Parameters should generally not be interacted with
        directly unless implementing new Layer subclasses or other core
        functionality. Wherever possible, users should interact with fittable
        parameters using Model- or Layer-level methods.

        Parameters
        ----------
        name : str
            This will also be the Parameter's key in `Phi._dict`.
        shape : N-tuple, default=()
            Shape of `Parameter.values`.
        prior : nems.distributions.Distribution or None; optional
            Prior distribution for this parameter, with matching shape.
        bounds : 2-tuple or None; optional
            (minimum, maximum) values for the entries of `Parameter.values`.
        default_bounds : str, default='infinite'
            Determines behavior when `bounds=None`.
            If `'infinite'`  : set bounds to (-np.inf, np.inf)
            If `'percentile'`: set bounds to tails of `Parameter.prior`.
                (prior.percentile(0.0001), prior.percentile(0.9999))
        zero_to_epsilon : False
            If True, change 0-entries of bounds to machine epsilon for float32,
            to avoid division by 0. Essentially this makes bounds an open
            (or half-open) interval if it includes zero values. This flag should
            be set any time that `Parameter.values` would be used in division,
            log, etc.
        initial_value : str, scalar, or ndarray, default='mean'
            Determines initial entries of `Parameter.values`.
            If `'mean'`   : set values to `Parameter.prior.mean()`.
            If `'sample'` : set values to `Parameter.prior.sample()`.
            If scalar     : set values to `np.full(Parameter.shape, scalar)`.
            If ndarray    : set values to array (must match `Parameter.shape`).

        See also
        --------
        nems.models.base.Model
        nems.layers.base.Layer
        nems.layers.base.Phi

        """
        self.name = name
        self.shape = tuple(shape)
        self.size = 1
        for axis in shape:
            self.size *= axis

        # default to multivariate normal
        if prior is None:
            zero = np.zeros(shape=self.shape)
            one = np.ones(shape=self.shape)
            prior = Normal(mean=zero, sd=one)  
        self.prior = prior
        if prior.shape != self.shape:
            raise ValueError(
                "Parameter.shape != Parameter.prior.shape for...\n"
                f"Parameter:       {self.name}\n"
                f"Parameter.shape: {self.shape}\n"
                f"prior.shape:     {prior.shape}"
                )

        if bounds is None:
            # Set bounds based on `default_bounds`
            if default_bounds == 'percentile':
                bounds = (prior.percentile(0.0001), prior.percentile(0.9999))
            elif default_bounds == 'infinite':
                bounds = (-np.inf, np.inf)
            else:
                raise ValueError(
                    "Unrecognized default_bounds for...\n"
                    f"Parameter:      {self.name}\n"
                    f"default_bounds: {default_bounds}\n"
                    "Accepted values are 'percentile' or 'infinite'."
                    )
        else:
            bounds = tuple(bounds)

        if zero_to_epsilon:
            # Swap 0 with machine epsilon for float32
            eps = np.finfo(np.float32).eps
            lower, upper = bounds
            lower = eps if lower == 0 else lower
            upper = eps if upper == 0 else upper
            bounds = (lower, upper)

        self.bounds = bounds

        if isinstance(initial_value, str) and (initial_value == 'mean'):
            value = self.prior.mean()
        elif isinstance(initial_value, str) and (initial_value) == 'sample':
            value = self.prior.sample(bounds=self.bounds)
        elif np.isscalar(initial_value):
            value = np.full(self.shape, initial_value)
        elif isinstance(initial_value, np.ndarray):
            value = initial_value
        else:
            raise ValueError(
                "Unrecognized initial_value for...\n"
                f"Parameter:     {self.name}\n"
                f"initial_value: {initial_value}\n"
                "Accepted values are 'mean', 'sample', scalar, or ndarray."
                )

        self.initial_value = np.ravel(value)

        # Must be set by `Phi.add_parameter` for `Parameter.values` to work.
        self.phi = None  # Pointer to parent Phi instance.
        self.first_index = None  # Location of data within Phi.get_vector()
        self.last_index = None

        self.is_frozen = False
        self.is_permanent = False
    
    # TODO: any other tracking/upkeep that needs to happen with
    #       freezing/unfreezing, or is the flag sufficient?
    def freeze(self):
        """Use parameter values for evaluation only, do not optimize."""
        self.is_frozen = True

    def unfreeze(self):
        """Make parameter values optimizable."""
        if not self.is_permanent:
            self.is_frozen = False

    def make_permanent(self):
        """Set parameter to a fixed value. Do not unfreeze or resample."""
        self.freeze()
        self.is_permanent = True

    @property
    def is_fittable(self):
        """Alias property for negation of `Parameter.is_frozen`."""
        return not self.is_frozen

    @property
    def values(self):
        """Get corresponding parameter values stored by parent Phi."""
        values = self.phi._get_parameter_vector(self)
        return np.reshape(values, self.shape)

    def sample(self, n=1, inplace=False):
        """Get or set new values by sampling from `Parameter.prior`.

        Re-sampling is used to ensure that the final sample falls within
        `Parameter.bounds`.
        
        Parameters
        ----------
        n : int
            Number of samples to get.
        inplace : bool
            If True, sampled values will be used to update `Parameter.values`
            inplace. Otherwise, the sampled values will be returned without
            changing current values.

        See also
        --------
        nems.layers.base.Phi.sample
        nems.distributions.base.Distribution.sample

        Return
        ------
        sample : ndarray

        """

        sample = self.prior.sample(n=n, bounds=self.bounds)
        if inplace:
            if self.is_permanent:
                # TODO: switch this to log.debug? or do something else maybe,
                #       instead of silently skipping.
                pass
            else:
                self.update(sample)
        else:
            return sample

    def mean(self, inplace=False):
        """Get, or set parameter values to, mean of `Parameter.prior`.
        
        Note that `Parameter.mean()` may return a value outside of
        `Parameter.bounds`. In that event, either priors or bounds should be
        changed.

        Parameters
        ----------
        inplace : bool, default=False
            If True, mean value will be used to update `Parameter.values`
            inplace Otherwise, mean will be returned without changing current
            values.

        Return
        ------
        mean : ndarray

        """
        mean = self.prior.mean()
        if inplace:
            if self.is_permanent:
                # TODO: switch this to log.debug? or do something else maybe,
                #       instead of silently skipping.
                pass
            else:
                self.update(mean)
        else:
            return mean

    def update(self, value, ignore_bounds=False):
        """Set `Parameters.values` to `value` by updating `Phi._array`.
        
        Parameters
        ----------
        value : scalar or array-like
            New value for `Parameter.values`. Must match `Parameter.shape`.
        ignore_bounds : bool
            If True, ignore `Parameter.bounds` when updating. Otherwise,
            new values will be rejected if they are less than `bounds[0]` or
            greater than `bounds[1]`.
        
        """
        value = np.asarray(value)
        if not ignore_bounds:
            lower, upper = self.bounds
            if np.any(value < lower) or np.any(value > upper):
                raise ValueError(
                    f"value out-of-bounds for...\n"
                    f"Parameter: {self.name}\n"
                    f"Bounds:    {self.bounds}\n"
                    f"Value:     {value}"
                )

        if np.shape(value) != self.shape:
            raise ValueError(
                f"Parameter {self.name} requires shape {self.shape}, but "
                f"{value} has shape {np.shape(value)}"
            )
        else:
            flat_value = np.ravel(value)
            mask = self.phi._get_parameter_mask(self)
            self.phi._array[mask] = flat_value

    def get_bounds_vector(self, none_for_inf=True):
        """Get a list with one bounds copy per entry in `Parameter.values`.
        
        Parameters
        ----------
        none_for_inf : bool, default=True
            If True, replace (+/-)`np.inf` with None for compatibility with
            `scipy.optimize.minimize`.
        
        Returns
        -------
        bounds : list of 2-tuples
        
        """
        bounds = self.bounds
        if none_for_inf:
            lower, upper = bounds
            if np.isinf(lower):
                lower = None
            if np.isinf(upper):
                upper = None
            bounds = (lower, upper)
        return [bounds] * self.size

    # Add compatibility for saving to .json    
    def to_json(self):
        """Encode Parameter object as json. See `nems.tools.json`."""
        data = {
            'kwargs': {
                'name': self.name,
                'shape': self.shape,
                'prior': self.prior,
                'bounds': self.bounds
            },
            'is_frozen': self.is_frozen,
            'is_permanent': self.is_permanent
        }
        return data

    @classmethod
    def from_json(cls, json):
        """Decode Parameter object from json. See `nems.tools.json`."""
        p = cls(**json['kwargs'])
        if json['is_permanent']:
            p.make_permanent()
        elif json['is_frozen']:
            p.freeze()
        return p

    def __str__(self):
        dash_break = "-"*16 + "\n"
        string = f"Parameter(name={self.name}, shape={self.shape})\n"
        string += dash_break
        string += f".prior:     {self.prior}\n"
        string += f".bounds:    {self.bounds}\n"
        if self.is_permanent:
            string += ".is_permanent: True\n"
        else:
            string += f".is_frozen: {self.is_frozen}\n"
        string += ".values:\n"
        string += f"{self.values}\n"
        string += dash_break
        return string

    def __repr__(self):
        string = f"Parameter(name={self.name}, shape={self.shape})\n"
        return string

    def __eq__(self, other):
        if isinstance(other, Parameter):
            return self.to_json() == other.to_json()
        else:
            return NotImplemented


    # TODO: Maybe it would be better to leave this functionality out?
    #       I keep running into cases where this fails and end up using
    #       .values anyway.

    # Add compatibility with numpy ufuncs, len(), and other methods that
    # should point to `Parameter.values` instead of `Parameter`.
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Propagate numpy ufunc operations to `Parameter.values`.
        
        Convenience method to eliminate excessive references to
        `Parameter.values`.
        
        Notes
        -----
        Works with `@` but not `np.dot`.

        See also
        --------
        https://numpy.org/doc/stable/reference/ufuncs.html
        
        """
        f = getattr(ufunc, method)
        # replace Parameter objects with Parameter.values
        subbed_inputs = [
            x.values if isinstance(x, Parameter) else x
            for x in inputs
            ]
        output = f(*subbed_inputs, **kwargs)
        return output

    @property
    def T(self):
        # np.transpose() works without this, but have to add .T separately.
        return self.values.T

    # Pass through iteration and math operators to `Parameter.values`.
    def __len__(self):
        return self.values.__len__()
    def __iter__(self):
        return self.values.__iter__()
    def __add__(self, other):
        return self.values.__add__(other)
    def __sub__(self, other):
        return self.values.__sub__(other)
    def __mul__(self, other):
        return self.values.__mul__(other)
    def __matmul__(self, other):
        return self.values.__matmul__(other)
    def __truediv__(self, other):
        return self.values.__truediv__(other)
    def __floordiv__(self, other):
        return self.values.__floordiv__(other)
    def __mod__(self, other):
        return self.values.__mod__(other)
    def __divmod__(self, other):
        return self.values.__divmod__(other)



import numpy as np
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import Input

from ..base import Backend, FitResults
from .cost import get_cost


class TensorFlowBackend(Backend):

    def __init__(self, nems_model, data, eval_kwargs=None):
        # TODO: Remove this after issues with using float32 in scipy have
        #       been fixed.
        nems_model.set_dtype(np.float32)
        super().__init__(nems_model, data, eval_kwargs=eval_kwargs)

    def _build(self, data, eval_kwargs=None):
        """Build a TensorFlow Keras model that corresponds to a NEMS Model. 

        Parameters
        ----------
        data : DataSet
            Model inputs, outputs and targets.
        eval_kwargs : dict
            Keyword arguments for `nems_model.evaluate`.

        Returns
        -------
        tensorflow.keras.Model

        Notes
        -----
        This method will only work if all Layers in the NEMS Model have
        implemented `as_tensorflow_layer`.
        
        """
        # TODO: what backend options to accept?

        batch_size = eval_kwargs.get('batch_size', 0)
        if batch_size == 0:
            data = data.prepend_samples()
            batch_size = None
        elif batch_size is not None:
            raise NotImplementedError(
                "tf.tensordot is failing for multiple batches b/c the axis "
                "numbers shift. Need to fix that before this will work."
            )

        inputs = data.inputs

        # Convert inputs to TensorFlow format
        tf_input_dict = {}
        for k, v in inputs.items():
            # Skip trial/sample dimension when determining shape.
            tf_in = Input(shape=v.shape[1:], name=k, batch_size=batch_size,
                          dtype=self.nems_model.dtype)
            tf_input_dict[k] = tf_in
        unused_inputs = list(tf_input_dict.keys())

        # Pass through Keras functional API to map inputs & outputs.
        last_output = None
        tf_data = tf_input_dict.copy()  # Need to keep actual Inputs separate
        tf_kwargs = {}  # TODO, regularizer etc.
        for layer in self.nems_model.layers:
            # Get all `data` keys associated with Layer args and kwargs
            # TODO: how are Layers supposed to know which one is which?
            #       have to check the name?
            layer_map = layer.data_map
            all_data_keys = layer_map.args + list(layer_map.kwargs.values())
            all_data_keys = np.array(all_data_keys).flatten().tolist()

            layer_inputs = []
            for k in all_data_keys:
                if k is None:
                    # Add last output
                    layer_inputs.append(last_output)
                else:
                    # Add Input with matching key
                    layer_inputs.append(tf_data[k])
                if k in unused_inputs:
                    unused_inputs.pop(unused_inputs.index(k))

            # Construct TF layer, provide input_shape for Layers that need that
            # extra information. If inputs is a singleton, unwrap it. Otherwise,
            # layers that expect a single input can break.
            input_shape = [keras.backend.int_shape(x) for x in layer_inputs]
            if len(layer_inputs) == 1:
                layer_inputs = layer_inputs[0]
                input_shape = input_shape[0]

            tf_layer = layer.as_tensorflow_layer(
                input_shape=input_shape, **tf_kwargs
                )
            last_output = tf_layer(layer_inputs)
            
            if isinstance(last_output, (list, tuple)):
                tf_data.update(
                    {k: v for k, v in zip(layer_map.out, last_output)
                    if k is not None}  # indicates unsaved intermediate output
                    )
            elif layer_map.out[0] is not None:
                tf_data[layer_map.out[0]] = last_output

        # Don't include inputs that were never actually passed to any Layers.
        tf_inputs = [v for k, v in tf_input_dict.items() if k not in unused_inputs]
        # For outputs, get all data entries that aren't inputs
        tf_outputs = [v for k, v in tf_data.items() if k not in tf_input_dict]

        model = tf.keras.Model(inputs=tf_inputs, outputs=tf_outputs,
                               name=self.nems_model.name)

        print('TF model built...')
        print(model.summary())

        return model

    def _fit(self, data, eval_kwargs=None, cost_function='squared_error',
             epochs=1, learning_rate=0.001, early_stopping_delay=100,
             early_stopping_patience=150, early_stopping_tolerance=5e-4,
             validation_split=0.0, validation_data=None):
        """Optimize `TensorFlowBackend.nems_model` using Adam SGD.
        
        Currently the use of other TensorFlow optimizers is not exposed as an
        option, but that may be added at a later time.

        TODO: allow loss functions other than mean squared error.

        Parameters
        ----------
        data : DataSet
            Model inputs, outputs and targets. Data must have shape (S, T, ...)
            where S is the number of samples/trials/etc, even if `S=1`.
        eval_kwargs : dict
            Keyword arguments for `nems_model.evaluate`.
        cost_function : str or function; default='squared_error'
            Specifies which metric to use for computing error while fitting.
            Uses mean squared error by default.
            If str      : Replace with `get_loss(str)`.
            If function : Use this function to compute errors. Should accept
                          two array arguments and return float.
        epochs : int
            Number of optimization iterations to perform.
        learning_rate : float; default=0.001.
            See docs for `tensorflow.keras.optimizers.Adam`.
        early_stopping_delay : int
            Minimum epoch before early stopping criteria are used. Set
            delay to 0 to use early stopping immediately.
        early_stopping_tolerance : float
            Minimum change in error between epochs considered to be an
            improvement (`min_delta` for `tf.keras.callbacks.EarlyStopping`).
            To disable early stopping, set tolerance to 0.
        early_stopping_patience : int
            Number of epochs to continue optimization for without improvement.
            (`patience` for `tf.keras.callbacks.EarlyStopping`.)
        validation_split : float
            Proportion of data to treat as temporary validation data
            (`validation_split` for `tf.keras.Model.fit`).
        validation_data : tuple of ndarray or tensors; optional.
            Specify validation data manually (overrides `validation_split`).
            See `tf.keras.Model.fit` for details.

        Returns
        -------
        FitResults

        """

        batch_size = eval_kwargs.get('batch_size', 0)
        if batch_size == 0:
            data = data.prepend_samples()
            batch_size = None

        # Replace cost_function name with function object.
        if isinstance(cost_function, str):
            print(f"cost_function: {cost_function}")
            cost_function = get_cost(cost_function)

        # TODO: support more keys in `fitter_options`.
        # Store initial parameters, compile optimizer and loss for model.
        initial_parameters = self.nems_model.get_parameter_vector()
        final_layer = self.nems_model.layers[-1].name
        self.model.compile(
            optimizer=keras.optimizers.Adam(learning_rate=learning_rate),
            loss={final_layer: cost_function}
        )

        # Build callbacks for early stopping, ... (what else?)
        callbacks = []
        loss_name = 'loss'
        if (validation_split > 0) or (validation_data is not None):
            loss_name = 'val_loss'
        if early_stopping_tolerance != 0:
            early_stopping = DelayedStopper(
                monitor=loss_name, patience=early_stopping_patience,
                min_delta=early_stopping_tolerance, verbose=1,
                restore_best_weights=True, start_epoch=early_stopping_delay,
                )
            callbacks.append(early_stopping)
        # Change back to default None if no callbacks were specified
        if len(callbacks) == 0: callbacks = None

        # TODO: This assumes a single output (like our usual models).
        #       Need to tweak this to be able to fit outputs from multiple
        #       layers. _build would need to establish a mapping I guess, since
        #       it has the information about which layer generates which output.
        inputs = data.inputs
        if len(data.targets) > 1:
            raise NotImplementedError("Only one target supported currently.")
        target = list(data.targets.values())[0]

        loss_fn = self.model.loss[final_layer]
        initial_error = loss_fn(
            tf.constant(target, dtype=tf.float32),
            tf.constant(self.model.predict(inputs), dtype=tf.float32)
            ).numpy()
        print(f"Initialial loss: {initial_error:.5f}")
        
        history = self.model.fit(
            inputs, {final_layer: target}, epochs=epochs,
            validation_split=validation_split, callbacks=callbacks,
            validation_data=validation_data
        )

        # Save weights back to NEMS model
        # Skip input layers.
        # TODO: This assumes there aren't any other extra layers added
        #       by TF. That might not be the case for some model types. Better
        #       approach would be to track keys for the specific layers that
        #       have the parameters we need.
        tf_model_layers = [
            layer for layer in self.model.layers
            if not layer.name in [x.name for x in self.model.inputs]
            ]
        layer_iter = zip(self.nems_model.layers, tf_model_layers)
        for nems_layer, tf_layer in layer_iter:
            nems_layer.set_parameter_values(tf_layer.weights_to_values())

        final_parameters = self.nems_model.get_parameter_vector()
        final_error = history.history[loss_name][-1]
        nems_fit_results = FitResults(
            initial_parameters, final_parameters, initial_error, final_error,
            backend_name='TensorFlow',
            misc={'TensorFlow History': history.history}
        )

        return nems_fit_results

    def predict(self, input, batch_size=0, **eval_kwargs):
        """Get output of `TensorFlowBackend.model` given `input`.
        
        Parameters
        ----------
        input : ndarray or list of ndarray, tensor or list of tensor.
        batch_size : int or None; default=0.
        eval_kwargs : dict
            Additional keyword arguments for `Model.evaluate`. Silently ignored.

        Returns
        -------
        np.ndarray
            Outpt of the associated Keras model.

        """
        if batch_size == 0:
            # Prepend samples
            if isinstance(input, (np.ndarray)):
                input = input[np.newaxis, ...]
                batch_size = None
            elif isinstance(input, list) and isinstance(input[0], np.ndarray):
                input = [x[np.newaxis, ...] for x in input]
                batch_size = None

        # TODO: Any kwargs needed here?
        return self.model.predict(input, batch_size=batch_size)


class DelayedStopper(tf.keras.callbacks.EarlyStopping):
    """Early stopper that waits before kicking in."""
    def __init__(self, start_epoch=100, **kwargs):
        super(DelayedStopper, self).__init__(**kwargs)
        self.start_epoch = start_epoch

    def on_epoch_end(self, epoch, logs=None):
        if epoch > self.start_epoch:
            super().on_epoch_end(epoch, logs)



"""Extend JSON encoding/decoding to work with `numpy.ndarray` and NEMS objects.

When encoding `obj` with `name = type(obj).__name__`, if an `encode_{name}`
method is found then that method will be used. Otherwise, the json package's
default encoding method will be used (`json.JSONEncoder.default`).

Warnings
--------
Subclasses of NEMS base classes should avoid overwriting `to_json` and
`from_json` whenever possible. NEMSEncoder will use the `to_json` method of a
subclass if overwritten, but NEMSDecoder will only ever use the `from_json`
method of the base class, which could cause confusion and/or mismatch between
methods.

If a subclass truly needs customized to/from methods, there are two options:
1) Corresponding `encode_{SubClassName}` and `decode_{SubClassName}` methods
   should also be added to NEMSEncoder and NEMSDecoder, respectively, so that
   the correct `from_json` method can be found by the decoder. Alternatively,
   add the imported subclass to the top-level `_NEMS_classes_to_encode` variable.
2) The base class `from_json` method should make use of a the subclass name
   (either from the default '__nems_json_subtype__' key or a defined alias) to
   re-direct `BaseClass.from_json(json)` to `SubClass.from_json(json)`.
   See `nems.layers.base.Layer.from_json` for an example.

Notes
-----
With the exception of objects from external libraries like `np.ndarray`,
`encode_{name}` and `decode_{name}` methods should simply reference
`obj.to_json()` or `cls.from_json(obj)` (along with, of course, defining the
`to_json` and `from_json` methods for that object's base class). The simplest
way to accomplish this is to add the imported class to `_NEMS_classes_to_encode`.

This ensures that the JSON encoder/decoder code does not depend on details of
individual classes, as they may change. Similarly, individual classes don't need
to worry about encoding/decoding details: they just need to return/load from a
dictionary that is suitable for standard `json.dumps`, i.e. it contains only
strings, lists, ints, etc. Additionally, classes don't need to worry about
encoding of other nems objects. I.e. a Phi instance contains a dict with
Parameter instances; however, since both classes implement `to_json`,
`json.dumps` will happily handle the details of encoding all Parameters before
storing them in the encoded Phi instance.

References
----------
.. [1] https://mathspp.com/blog/custom-json-encoder-and-decoder

Examples
--------
To encode data stored in a `numpy.ndarray` as JSON:
>>> my_data = np.arange(1, 5)
>>> encoded_data = json.dumps(my_data, cls=NEMSEncoder)
>>> encoded_data
'{"data": [1, 2, 3, 4], "dtype": "int32", "shape": [4],
    "__nems_json_type__": "ndarray"}'

>>> decoded_data = json.loads(encoded_data, cls=NEMSDecoder)
>>> decoded_data
array([1, 2, 3, 4])

"""

import os
import re
import json
import inspect
from functools import partialmethod

import numpy as np

from nems.distributions.base import Distribution
from nems.layers.base import Layer, Phi, Parameter
from nems.models.base import Model
from nems.backends.base import FitResults


_NEMS_classes_to_encode = [
    Distribution, Layer, Phi, Parameter, Model, FitResults
    ]


class NEMSEncoder(json.JSONEncoder):
    """An extended JSON encoder. See `help(nems.tools.json)` for docs."""

    def default(self, obj):
        # Iterate through object's class, immediate parent class(es), then
        # grandparent class(es), and so on. Stop when one of these class names
        # has a corresponding `encode_{name}` method.
        # (This means NEMS objects can subclassed as many times as desired, and
        #  this method will still find the correct base class).
        inheritance_list = inspect.getmro(type(obj))
        for cls in inheritance_list:
            encoder = getattr(self, f"encode_{cls.__name__}", None)
            if encoder is not None:
                encoded = encoder(obj)
                encoded["__nems_json_type__"] = cls.__name__
                encoded["__nems_json_subtype__"] = inheritance_list[0].__name__
                return encoded
        else:
            super().default(obj)

    def encode_ndarray(self, array):
        data = array.tolist()
        return {'data': data, 'dtype': str(array.dtype), 'shape': array.shape}


class NEMSDecoder(json.JSONDecoder):
    """An extended JSON decoder. See `help(nems.tools.json)` for docs."""

    def __init__(self, **kwargs):
        kwargs["object_hook"] = self.object_hook
        super().__init__(**kwargs)

    def object_hook(self, obj):
        try:
            name = obj["__nems_json_type__"]
            decoder = getattr(self, f"decode_{name}")
        except (KeyError, AttributeError):
            return obj
        else:
            return decoder(obj)

    def decode_ndarray(self, obj):
        return np.asarray(obj['data'], dtype=obj['dtype']).reshape(obj['shape'])


# Monkey-patch generic encode/decode methods for listed NEMS classes.
@staticmethod
def _generic_NEMS_encoder(obj):
    return obj.to_json()
@staticmethod
def _generic_NEMS_decoder(obj, cls=None):
    return cls.from_json(obj)

for cls in _NEMS_classes_to_encode:
    name = cls.__name__
    decoder = partialmethod(_generic_NEMS_decoder, cls=cls)
    setattr(NEMSEncoder, f'encode_{name}', _generic_NEMS_encoder)
    setattr(NEMSDecoder, f'decode_{name}', decoder)


# Convenience functions so that front-end users don't need to remember to use
# cls=NEMSEncoder/Decoder.
def nems_to_json(obj):
    """Encode `obj` as a json string. Supports ndarray and NEMS objects."""
    return json.dumps(obj, cls=NEMSEncoder)

def nems_from_json(obj):
    """Decode json `obj`. Supports ndarray and NEMS objects."""
    return json.loads(obj, cls=NEMSDecoder)


def save_model(model, filepath):
    """Save a Model to `filepath` as a json file.

    TODO: refactor using pathlib, options to use different permissions.

    Parameters
    ----------
    model: nems.models.base.Model
        NEMS model object to save.
    filepath : string
        Full path including file name and parent directories.
    
    Notes
    -----
    Currently this function is hard-coded to set broad permissions for
    the saved file.

    """
    data = nems_to_json(model)
    dirpath = os.path.dirname(filepath)

    if not os.path.exists(dirpath):
        os.makedirs(dirpath, mode=0o0777)
    with open(filepath, mode='w+') as f:
        f.write(data)
        f.close()
        try:
            os.chmod(filepath, 0o666)
        except PermissionError:
            # File should already exist with the correct permissions
            pass


def load_model(filepath):
    """Load a Model from a json file saved at `filepath`.

    Parameters
    ----------
    filepath : string
        Full path to the location of the saved Model, including file name.

    Returns
    -------
    model : nems.models.base.Model
        A NEMS Model object.

    """
    with open(filepath, mode='r') as f:
        data = f.read()

    model = nems_from_json(data)

    return model


def generate_model_filepath(model, basepath="", max_length=100):
    """Get filepath for saving Model based on `Model.name`.
    
    Uses `Model.name` to guess an appropriate directory name. Directory name is
    filtered for invalid characters and truncated with a hash if the name is
    longer than `max_length`. The final filepath is returned in the form of:
    `'<basepath>/<directory name>/model.json'`.

    Parameters
    ----------
    model : nems.Model
    basepath : str; default="".
        Path where save directory should be created.
    max_length : int >= 22; default=100.
        Maximum number of characters to include in the created directory name.
        If more this number is exceeded, the final 22 characters will be
        replaced with '...<19-character hash>'.
        NOTE: `max_length` does *not* include `basepath` or the 'model.json'
              suffix.
        
    Returns
    -------
    filepath : string

    """

    guess = model.name

    # Remove problematic characters
    guess = re.sub('[/]', '_', guess)
    guess = re.sub('[:]', '', guess)
    guess = re.sub('[,]', '', guess)

    if len(guess) >= max_length:
        # If the total filepath is too long, it causes filesystem errors.
        guess = guess[:max_length-22] + '...' + str(hash(guess))

    return os.path.join(basepath, guess, 'model.json')



import os, sys


# https://stackoverflow.com/questions/8391411/how-to-block-calls-to-print
# Alexander C
class PrintingBlocked:
    """Context manager for temporarily blocking `sys.stdout`.
    
    Examples
    --------
    >>> def test(i):
    >>>     print(f'{i} not blocked')
    >>>
    >>> with PrintingBlocked():
    >>>     test(1)
    >>> test(2)
    2 not blocked
    
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


# https://stackoverflow.com/questions/3160699/python-progress-bar
# imbr
def progress_bar(iterable, prefix="", character='#', size=60, out=sys.stdout):
    """Simple text-based progress bar.

    Parameters
    ----------
    iterable : iterable.
    prefix : str; default="".
        Fixed text header placed to the left of the progress bar.
    character : str; default='#'.
        Single-character string to represent progress increments.
    size : int; default=60.
        Number of characters to use to represent the progress bar.
    out : I/O stream; default=sys.stdout.
    
    Examples
    --------
    >>> import time    
    >>> for i in progressbar(range(15), "Computing: ", 40):
    >>>     time.sleep(0.1)  # any code you need
    
    """
    count = len(iterable)
    def show(j):
        x = int(size*j/count)
        print(f"{prefix}[{character*x}{('.'*(size-x))}] {j}/{count}", end='\r',
              file=out, flush=True)
    show(0)
    for i, item in enumerate(iterable):
        yield item
        show(i+1)
    print("\n", flush=True, file=out)



from scipy import stats
import numpy as np

from .base import Distribution


class HalfNormal(Distribution):
    """HalfNormal prior.

    Parameters
    ----------
    sigma : scalar or ndarray
        Standard deviation of distribution
    
    """

    def __init__(self, sd):
        self._sd = np.asarray(sd)
        self.distribution = stats.halfnorm(scale=self._sd)

    def __repr__(self):
        sd = self.value_to_string(self._sd)
        return 'HalfNormal(σ={})'.format(sd)



import numpy as np


def indices_by_fraction(data, fraction=0.9, axis=0):
    """Get indices that split `data` into two parts along one axis.

    Parameters
    ----------
    data : numpy.ndarray
        An array to be split.
    fraction : float in range [0,1]
        The proportion of data that will go into the "left" split. The proportion
        in the "right" split will be `1 - fraction`, or as close as possible
        after truncating to the nearest integer bin index.
    axis : int
        The axis on which to perform the split.
    

    Returns
    -------
        2-tuple of numpy.ndarray
            The indices corresponding to before and after the specified
            proportion of the data, suitable as arguments for `split_at_indices`.
    
    """

    split_idx = int(data.shape[axis] * fraction)
    before_fraction = np.arange(0, split_idx)
    after_fraction = np.arange(split_idx, data.shape[axis])

    return before_fraction, after_fraction


def split_at_indices(data, idx1, idx2, axis=0):
    """Split `data` into two parts along one axis by taking slices."""

    first_subset = np.take(data, idx1, axis=axis)
    second_subset = np.take(data, idx2, axis=axis)

    return first_subset, second_subset


# TODO: should this return a generator instead of list?
#       would use less data if the number of indices gets really large
def get_jackknife_indices(data, n, axis=0, full_shuffle=False,
                          shuffle_jacks=True):
    """Generate indices for selecting `n` jackknife replicates from `data`.

    Parameters
    ----------
    data : numpy.ndarray
        Array to make jackknife replicates of.
    n : int
        Number of jackknife replicates to generate indices for.
    axis : int
        Axis on which replicate indices will be generated.
    full_shuffle : bool
        If true, shuffle all array entries along `axis`.
    shuffle_jacks : bool
        If true, shuffle the ordering of the replicate indices. This is useful
        if you want the jackknife replicates to be semi-random but still want
        to (mostly) maintain the original structure of the data.
        Redundant if setting `full_shuffle = True`.

    Returns
    -------
    jack_indices : list
        Each element in the list is a suitable `indices` argument for
        `get_jackknife`.
    
    """
    arrays = []
    indices = np.arange(0, data.shape[axis])
    if full_shuffle:
        indices = np.random.permutation(indices)
    jack_indices = np.array_split(indices, n)
    if shuffle_jacks:
        jack_indices = np.random.permutation(jack_indices)

    return jack_indices

def get_jackknife(data, indices, axis=0):
    """Get a jackknife replicate by deleting slices at `indices` along `axis`."""
    return np.delete(data, obj=indices, axis=axis)

# TODO: what other generic split functions would be useful here?



"""Base structures for representing data transformation steps of Models.

Layer     : Top-level component. Implements an `evaluate` method, which carries
            out the transformation.
Phi       : Stores and manages fittable parameters for the transformation.
Parameter : Low-level representation of individual parameters.

"""

from .layer import Layer
from .phi import Phi
from .parameter import Parameter
from .map import DataMap
from .errors import ShapeError



# This code is derived from the gammatone toolkit, licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING

"""Constructs sets of equivalent rectangular bandwidth gammatone filters."""

from __future__ import division

import numpy as np
from scipy import signal as sgn


def erb_space(low_freq=100.0, high_freq=44100/4, n=100):
    """Compute `n` uniformly spaced frequencies on an ERB scale.
    
    For a definition of ERB, see Moore, B. C. J., and Glasberg, B. R. (1983).
    "Suggested formulae for calculating auditory-filter bandwidths and
    excitation patterns," J. Acoust. Soc. Am. 74, 750-753.

    Parameters
    ----------
    low_freq : float; default=100.0.
    high_freq : float; deafult=44100/4.
    n : int; default=100.

    Returns
    -------
    np.ndarray, dtype=float.

    """
    return erb_point(low_freq, high_freq, np.arange(1, n+1)/n)


def erb_point(low_freq, high_freq, fraction, glasberg_moore=None):
    """Calculates point(s) on an ERB scale.
    
    If `fraction` is a scalar a single point is returned, but if `fraction` is
    an array then an array of points is returned. Each point will be between
    `low_freq` and `high_freq`, with the exact position determined by `fraction`.
    If `fraction == 1`, `low_freq` is used. If `fraction == 0`, `high_freq` is
    used.
    
    `fraction` can actually be outside the range `[0, 1]`, which in general
    isn't very meaningful, but might be useful when `fraction` is rounded a
    little above or below `[0, 1]` (eg. for plot axis labels).

    Parameters
    ----------
    low_freq : float.
    high_freq : float.
    fraction : float or np.ndarray.
    glasberg_moore : dict; optional.
        A dictionary with keys 'ear_q', 'min_bw', and 'order'. Change these
        values if you wish to use a different ERB scale.

        Defaults are Glasberg and Moore parameters:
        'ear_q': 9.26449
        'min_bw': 24.7
        'order': 1

    Returns
    -------
    float or np.ndarray.

    """

    if glasberg_moore is None: glasberg_moore = {}
    # Default to Glasberg and Moore parameters.
    ear_q = glasberg_moore.get('ear_q', 9.26449)
    min_bw = glasberg_moore.get('min_bw', 24.7)
    order = glasberg_moore.get('order', 1)

    # All of the following expressions are derived in Apple TR #35, "An
    # Efficient Implementation of the Patterson-Holdsworth Cochlear Filter
    # Bank." See pages 33-34.
    erb_point = (
        -ear_q * min_bw
        + np.exp(
            fraction * (
                -np.log(high_freq + ear_q * min_bw)
                + np.log(low_freq + ear_q * min_bw)
                )
        ) *
        (high_freq + ear_q * min_bw)
    )
    
    return erb_point


def centre_freqs(fs, num_freqs, cutoff, f_max=None):
    """Calculate centre frequencies for use by `make_erb_filters`.

    Parameters
    ----------
    fs : int.
        Sampling frequency.
    num_freqs : int.
        Number of centre frequencies to calculate.
    cutoff : float.
        Lower cutoff frequency.
    f_max : float; optional.
        Upper cutoff frequency. Added by SVD, 2020-04-09.

    Returns
    -------
    np.ndarray, dtype=float.

    """
    if f_max is None: f_max = fs/2
    return erb_space(cutoff, f_max, num_freqs)


def make_erb_filters(fs, centre_freqs, width=1.0, glasberg_moore=None):
    """Compute filter coefficients for a bank of Gammatone filters.
    
    These filters were defined by Patterson and Holdworth for simulating the
    cochlea. Each row of the filter arrays contains the coefficients for four
    second order filters. The transfer function for these four filters share the
    same denominator (poles) but have different numerators (zeros). All of these
    coefficients are assembled into one vector that `erb_filterbank` can use
    to implement the filter.

    Parameters
    ----------
    fs : int.
        Sampling frequency.
    centre_freqs : np.ndarray, dtype=float.
        Center frequencies returned by `centre_freqs(fs, ...)`.
    width : float; default=1.0.
        TODO: what does width affect?
    glasberg_moore : dict; optional.
        A dictionary with keys 'ear_q', 'min_bw', and 'order'. Change these
        values if you wish to use a different ERB scale.

        Defaults are Glasberg and Moore parameters:
        'ear_q': 9.26449
        'min_bw': 24.7
        'order': 1

    Returns
    -------
    np.ndarray, dtype=float.
    
    Notes
    -----
    This implementation fixes a problem in the original code by computing four
    separate second order filters. This avoids a big problem with round off
    errors in cases of very small cfs (100Hz) and large sample rates (44kHz).
    The problem is caused by roundoff error when a number of poles are combined,
    all very close to the unit circle. Small errors in the eigth order
    coefficient, are multiplied when the eigth root is taken to give the pole
    location. These small errors lead to poles outside the unit circle and
    instability. Thanks to Julius Smith for leading me to the proper explanation.
    
    Examples
    --------
    Evaluate the response of a 10-channel filterbank:
    >>> cf = centre_freqs(fs=16000, num_freqs=10, cutoff=100)
    >>> filter_coefficients = make_erb_filters(fs=16000, cf, width=1)
    >>> erb_filterbank(np.sin(np.arange(500)), filter_coefficients).shape
    (500, 10)
    
    Copyright
    ---------
    Rewritten by Malcolm Slaney@Interval.  June 11, 1998.
    (c) 1998 Interval Research Corporation
    (c) 2012 Jason Heeris (Python implementation)

    """
    T = 1 / fs

    if glasberg_moore is None: glasberg_moore = {}
    # Default to Glasberg and Moore parameters.
    ear_q = glasberg_moore.get('ear_q', 9.26449)
    min_bw = glasberg_moore.get('min_bw', 24.7)
    order = glasberg_moore.get('order', 1)

    erb = width*((centre_freqs/ear_q)**order + min_bw**order)**(1/order)
    B = 1.019 * 2 * np.pi * erb

    arg = 2 * centre_freqs * np.pi * T
    vec = np.exp(2j * arg)

    A0 = T
    A2 = 0
    B0 = 1
    B1 = -2 * np.cos(arg) / np.exp(B * T)
    B2 = np.exp(-2 * B * T)
    
    rt_pos = np.sqrt(3 + 2 ** 1.5)
    rt_neg = np.sqrt(3 - 2 ** 1.5)
    
    common = -T * np.exp(-(B * T))
    
    # TODO: This could be simplified to a matrix calculation involving the
    # constant first term and the alternating rt_pos/rt_neg and +/-1 second
    # terms
    k11 = np.cos(arg) + rt_pos * np.sin(arg)
    k12 = np.cos(arg) - rt_pos * np.sin(arg)
    k13 = np.cos(arg) + rt_neg * np.sin(arg)
    k14 = np.cos(arg) - rt_neg * np.sin(arg)

    A11 = common * k11
    A12 = common * k12
    A13 = common * k13
    A14 = common * k14

    gain_arg = np.exp(1j * arg - B * T)

    gain = np.abs(
            (vec - gain_arg * k11)
          * (vec - gain_arg * k12)
          * (vec - gain_arg * k13)
          * (vec - gain_arg * k14)
          * (  T * np.exp(B * T)
             / (-1 / np.exp(B * T) + 1 + vec * (1 - np.exp(B * T)))
            )**4
        )

    allfilts = np.ones_like(centre_freqs)
    
    fcoefs = np.column_stack([
        A0 * allfilts, A11, A12, A13, A14, A2*allfilts,
        B0 * allfilts, B1, B2,
        gain
    ])
    
    return fcoefs


def erb_filterbank(wave, coefs):
    """Process an input waveform with a gammatone filter bank.

    Parameters
    ----------
    wave : np.ndarray.
        Sound waveform with shape (T,) or (T,1).
    coefs : np.ndarray.
        Gammatone filter coefficients returned by `make_erb_filters`,
        with shape (`num_freqs`, 10)
    
    Returns
    -------
    np.ndarray
        Filter outputs, with shape (T, `num_freqs`).

    Copyright
    ---------
    Malcolm Slaney @ Interval, June 11, 1998.
    (c) 1998 Interval Research Corporation
    Thanks to Alain de Cheveigne' for his suggestions and improvements.
    (c) 2013 Jason Heeris (Python implementation)

    """

    # TODO: why the hard-coded 9?
    output = np.zeros((coefs[:,9].shape[0], wave.shape[0]))
    
    gain = coefs[:, 9]
    # A0, A11, A2
    As1 = coefs[:, (0, 1, 5)]
    # A0, A12, A2
    As2 = coefs[:, (0, 2, 5)]
    # A0, A13, A2
    As3 = coefs[:, (0, 3, 5)]
    # A0, A14, A2
    As4 = coefs[:, (0, 4, 5)]
    # B0, B1, B2
    Bs = coefs[:, 6:9]
    
    # Loop over channels
    for idx in range(coefs.shape[0]):
        # These seem to be reversed (in the sense of A/B order), but that's what
        # the original code did...
        # Replacing these with polynomial multiplications reduces both accuracy
        # and speed.
        y1 = sgn.lfilter(As1[idx], Bs[idx], wave)
        y2 = sgn.lfilter(As2[idx], Bs[idx], y1)
        y3 = sgn.lfilter(As3[idx], Bs[idx], y2)
        y4 = sgn.lfilter(As4[idx], Bs[idx], y3)
        output[idx, :] = y4 / gain[idx]
        
    return output.T  # Transpose to conform w/ NEMS data format



import math

import numpy as np


def raster_to_spike_times(raster, fs):
    """Convert spike raster to a list of spike time arrays.
    
    Parameters
    ----------
    raster : np.ndarray.
        Binary spike raster of shape (T, N), where T is the number of time bins
        and N is the number of neurons.
    fs : int.
        Sampling rate of `raster`. Determines precision of spike times.

    Returns
    -------
    spike_times_in_seconds : list of ndarray.
    
    """
    spike_times_in_seconds = [
        raster[..., i].nonzero()[0] / fs
        for i in range(raster.shape[1])
        ]

    return spike_times_in_seconds


def spike_times_to_raster(spike_times, fs, duration=None):
    """Convert nested list of spike times to a spike raster.

    TODO: Add option to convert times from other units (i.e conversion factor
          similar to plotting utility).

    TODO: This is my first-pass naive implementation. There are likely some
          details I'm not considering, ask Stephen to revise.

    Parameters
    ----------
    spike_times : list of list, or list of np.ndarray.
        Each list element should represent all spike times in seconds for one
        neuron, in order of increasing time.
    fs : int.
        Sampling rate for spike raster.
    duration : float.
        Amount of time in seconds that the raster should span. If not specified,
        duration will be determined by the largest spike time.

    Returns
    -------
    raster : np.ndarray, dtype=bool.
    
    """
    if duration is not None:
        n_time_bins = math.ceil(duration*fs)
    else:
        last_spike_time = max([times[-1] for times in spike_times])
        n_time_bins = math.ceil(last_spike_time*fs)

    raster_columns = []
    duplicate_spikes = 0
    truncated_spikes = 0
    for i, times in enumerate(spike_times):
        column = np.zeros(shape=(n_time_bins, 1), dtype=np.bool)

        for j, t in enumerate(times):
            spike_bin = round(t*fs)
            if spike_bin == n_time_bins:
                # Exactly at end of duration, round down instead of up.
                spike_bin -= 1
            elif spike_bin > n_time_bins:
                # Spike times past duration, truncate.
                truncated_spikes += len(times[j:])
                break

            # NOTE: Multiple spikes in the same bin will be overwritten.
            if column[spike_bin, 0] == 1:
                # Already a spike here, log the count for debugging.
                duplicate_spikes += 1
            else:
                column[spike_bin, 0] = 1

        raster_columns.append(column)

    print(f'{duplicate_spikes} spikes dropped due to bin number rounding.')
    print(f'{truncated_spikes} spikes dropped due to truncated duration.')
    raster = np.concatenate(raster_columns, axis=1)

    return raster



from .base import Model
from nems.layers import (
    WeightChannels, WeightChannelsGaussian, FiniteImpulseResponse,
    RectifiedLinear, DoubleExponential
    )

class LN_STRF(Model):

    def __init__(self, channels, time_bins, rank=None,
                 gaussian=True, nonlinearity='DoubleExponential',
                 nl_kwargs=None, **model_init_kwargs):
        """Linear-nonlinear Spectro-Temporal Receptive Field model.

        Contains the following layers:
        1. WeightChannels, WeightChannelsGaussian, or None (if full rank).
        2. FiniteImpulseResponse
        3. DoubleExponential, RectifiedLinear, or another static nonlinearity.

        Expects a single sound spectrogram as input, with shape (T, N), and a
        single recorded neural response as a target, with shape (T, 1), where
        T and N are the number of time bins and spectral channels, respectively.

        Based on model architectures described in:
        Thorson, Lienard and David (2015)
        doi: 10.1371/journal.pcbi.1004628

        Parameters
        ----------
        channels : int.
            Number of spectral channels in spectrogram.
        time_bins : int.
            Number of "taps" in FIR filter. We have found that a 150-250ms filter
            typically sufficient for modeling A1 responses, or 15-25 bins for
            a 100 Hz sampling rate.
        rank : int; optional.
            Number of spectral weightings used as input to a reduced-rank filter.
            For example, `rank=1` indicates a frequency-time separable STRF.
            If unspecified, a full-rank filter will be used.
        gaussian : bool; default=True.
            If True, use gaussian functions (1 per `rank`) to parameterize
            spectral weightings. Unused if `rank is None`.
        nonlinearity : str; default='DoubleExponential'.
            Specifies which static nonlinearity to apply after the STRF.
            Default is the double exponential nonlinearity used in the paper
            cited above.
        nl_kwargs : dict; optional.
            Additional keyword arguments for the nonlinearity Layer, like
            `no_shift` or `no_offset` for `RectifiedLinear`.
        model_init_kwargs : dict; optional.
            Additional keyword arguments for `Model.__init__`, like `dtype`
            or `meta`.

        """

        super().__init__(**model_init_kwargs)

        # Add STRF
        if rank is None:
            # Full-rank finite impulse response
            fir = FiniteImpulseResponse(shape=(time_bins, channels))
            self.add_layers(fir)
        else:
            wc_class = WeightChannelsGaussian if gaussian else WeightChannels
            wc = wc_class(shape=(channels, rank))
            fir = FiniteImpulseResponse(shape=(time_bins, rank))
            self.add_layers(wc, fir)

        # Add static nonlinearity
        if nonlinearity in ['DoubleExponential', 'dexp', 'DEXP']:
            nl_class = DoubleExponential
        elif nonlinearity in ['RectifiedLinear', 'relu', 'ReLU']:
            nl_class = RectifiedLinear
        else:
            raise ValueError(
                f'Unrecognized nonlinearity for LN model:  {nonlinearity}.')
        
        if nl_kwargs is None: nl_kwargs = {}
        nonlinearity = nl_class(shape=(1,), **nl_kwargs)
        self.add_layers(nonlinearity)


    @classmethod
    def from_data(cls, input, filter_duration, sampling_rate=1000, **kwargs):
        channels = input.shape[-1]
        time_bins = int(filter_duration/1000 * sampling_rate)
        # TODO: modify initial parameters based on stimulus statistics?
        return LN_STRF(channels, time_bins, **kwargs)

    def fit_LBHB():
        # TODO: 3-stage fit with freezing/unfreezing NL
        pass

    # TODO
    # @module('LNSTRF')
    def from_keyword(keyword):
        # Return a list of module instances matching this pre-built Model?
        # That way these models can be used with kw system as well, e.g.
        # model = Model.from_keywords('LNSTRF')
        #
        # But would need the .from_keywords method to check for list vs single
        # module returned.
        pass



"""Tools for debugging complex routines like model fitting."""

from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.gridspec import GridSpec
import numpy as np

from nems.visualization import (
    prediction_vs_target, iteration_vs_error, evals_per_iteration,
    parameter_space_pca, parameter_pca_table, scipy_gradient
    )
from .io import PrintingBlocked, progress_bar
from .json import save_model, load_model


def get_model_fits(model, input, target, iterations=None, save_path=None,
                   load_path=None, backend='scipy', fitter_options=None, 
                   **fit_kwargs):
    """Generate a list of Models by fitting one iteration at a time.
    
    TODO: Truncate iterations if fit is terminating b/c of tolerance criteria.
    TODO: Implement `iterations=None` (i.e. fit until stop critiera are met).

    Parameters
    ----------
    model : nems.Model.
        Initial model to base fits on.
    input : np.ndarray, dict, or DataSet.
        See `nems.models.base.Model.evaluate`.
    target : np.ndarray or dict.
        See `nems.models.base.Model.fit`.
    iterations : int; optional.
        Number of fit iterations to record. For `backend='scipy'` this is
        interpreted as `{'options': {'maxiter': iterations}}`. For `backend='tf'`
        this is interpreted as `epochs=iterations`.
    save_path : str or Path; optional.
        Path where models should be saved after each fit iteration. Must point
        to a directory.
    load_path : str or Path; optional.
        Path to load previously fit models from. Must point to a directory
        containing NEMS models saved to json format.
    backend : str; default='scipy'.
        Name of backend to use for fitting. See `nems.models.base.Model.fit`.
    fitter_options : dict; optional.
        See `nems.models.base.Model.fit` and backend documentation.
    fit_kwargs : dict; optional.
        Additional keyword arguments for `nems.models.base.Model.fit`.

    Returns
    -------
    list of nems.Model.
        Length iterations + 1.

    See also
    --------
    nems.models.base.Model
    nems.tools.json.save_model
    nems.tools.json.load_model
    
    """

    if load_path is not None:
        # TODO: How to ensure these are loaded in the correct order?
        #       Currently assumes they're ordered numerically, increasing.
        directory = Path(load_path).glob('**/*')
        models = [load_model(f) for f in directory if f.is_file()]
        return models[:iterations+1]

    if fitter_options is None: fitter_options = {}

    # Overwrite `iterations` if specified in `fitter_options`,
    # always fit one iteration at a time.
    if backend == 'scipy':
        if 'options' in fitter_options:
            iterations = fitter_options['options'].get('maxiter', iterations)
        else:
            fitter_options['options'] = {}
        fitter_options['options']['maxiter'] = 1
    elif backend == 'tf':
        iterations = fitter_options.get('epochs', iterations)
        fitter_options['epochs'] = 1
    else:
        raise NotImplementedError(
            f'{backend} backend not implemented for `get_model_fits`.'
            )

    models = [model]
    if iterations is None: raise NotImplementedError(
        "Must specify iterations in advance for now."
    )
    for _ in progress_bar(range(iterations), prefix="Fitting iteration: "):
        with PrintingBlocked():
            fitted_model = models[-1].fit(
                input, target, fitter_options=fitter_options, backend=backend,
                **fit_kwargs
                )
            models.append(fitted_model)

    if save_path is not None:
        directory = Path(save_path)
        for i, m in enumerate(models):
            p = directory / f'{i}.json'
            save_model(m, p)

    return models


def debug_plot(models, input, target, sampling_rate=None, xlim=None,
               figsize=None):
    """Plot error, gradient, parameters, and prediction for each Model.
    
    Parameters
    ----------
    models : list of nems.Model
        Model.results must not be None for all but the first Model, and all
        models must have the same layers and hyperparameters.
        For example, as returned by `get_model_fits`.
    input : np.ndarray, dict, or DataSet.
        See `nems.models.base.Model.evaluate`.
    target : np.ndarray or dict.
        See `nems.models.base.Model.fit`.
    sampling_rate : float; optional.
        If given, label time-axis of some plots with real units (seconds
        by default). See `nems.visualization.tools.ax_bins_to_seconds`.
    xlim : tuple; optional.
        Positional arguments for: `ax.set_xlim(*xlim)` for prediction plot.
        If `sampling_rate` is not None, specified in units of seconds instead
        of bins.
    figsize : 2-tuple of int; optional.
         Figure size in inches, (width, height).

    Returns
    -------
    fig : matplotlib.pyplot.figure
    
    """

    if figsize is None: figsize = (10, 10)
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    grid = GridSpec(12, 12, figure=fig)

    #                         row,   col
    a1 = fig.add_subplot(grid[ :3,   :7])   # top left timeseries
    a2 = fig.add_subplot(grid[3:5,   :7])   # top left time series 2
    a3 = fig.add_subplot(grid[ :5,  7: ])   # top right scatter
    a4 = fig.add_subplot(grid[6:8,   :7])   # middle left timeseries
    a5 = fig.add_subplot(grid[6:8,  7: ])   # middle right table
    a6 = fig.add_subplot(grid[8:12,  : ])   # bottom timeseries

    # Plot iteration vs error, number of fn evaluations.
    # Wrap with `make_axes_plotter` to standardize axes formatting
    # (remove upper and right box border, convert bins to time, etc)
    iteration_vs_error(models, ax=a1)
    if models[1].results.backend == 'SciPy':
        # Number of cost function evals vs iterations
        evals_per_iteration(models, ax=a2)
        # Norm of gradient vs iterations
        scipy_gradient(models, ax=a4)

    # Visualize path of optimization through parameter space using PCA
    parameter_space_pca(models, ax=a3)

    # TODO: Can't get this to work right with gridspec
    # Table of first 5 PCs and parameter with largest weight for each PC
    # parameter_pca_table(models, n=5, m=1, fontsize=12, ax=a5)
    a5.axis('off')

    if (xlim is not None) and (sampling_rate is not None):
        # Scale xlim from units of seconds to units of bins
        xmin, xmax = xlim
        xlim = (xmin*sampling_rate, xmax*sampling_rate)

    # Plot predicted vs actual response, input heatmap
    prediction_vs_target(
        input, target, models[0], ax=a6, title='Model prediction',
        show_input=True, sampling_rate=sampling_rate, xlim=xlim
        )

    return fig


def debug_fitter(model, input, target, iterations=None, sampling_rate=None,
                 xlim=None, figsize=None, save_path=None, load_path=None,
                 backend='scipy', fitter_options=None, **fit_kwargs):
    """Plot error, gradient, parameters, and prediction for each fit iteration.

    Parameters
    ----------
    model : nems.Model.
        Initial model to base fits on.
    input : np.ndarray, dict, or DataSet.
        See `nems.models.base.Model.evaluate`.
    target : np.ndarray or dict.
        See `nems.models.base.Model.fit`.
    iterations : int; optional.
        Number of fit iterations to record. For `backend='scipy'` this is
        interpreted as `{'options': {'maxiter': iterations}}`. For `backend='tf'`
        this is interpreted as `epochs=iterations`.
    sampling_rate : float; optional.
        If given, label time-axis of some plots with real units (seconds
        by default). See `nems.visualization.tools.ax_bins_to_seconds`.
    xlim : tuple; optional.
        Positional arguments for: `ax.set_xlim(*xlim)` for prediction plot.
        If `sampling_rate` is not None, specified in units of seconds instead
        of bins.
    figsize : 2-tuple of int; optional.
         Figure size in inches, (width, height).
    save_path : str or Path; optional.
        Path where models should be saved after each fit iteration. Must point
        to a directory.
    load_path : str or Path; optional.
        Path to load previously fit models from. Must point to a directory
        containing NEMS models saved to json format.
    backend : str; default='scipy'.
        Name of backend to use for fitting. See `nems.models.base.Model.fit`.
    fitter_options : dict; optional.
        See `nems.models.base.Model.fit` and backend documentation.
    fit_kwargs : dict; optional.
        Additional keyword arguments for `nems.models.base.Model.fit`.

    Returns
    -------
    matplotlib.pyplot.figure
    list of nems.Model
        Length iterations + 1.
    
    """

    # Get list of models, starting with initial model and appending one per
    # fit iteration.
    models = get_model_fits(
        model, input, target, iterations=iterations, backend=backend,
        fitter_options=fitter_options, save_path=save_path, load_path=load_path,
        **fit_kwargs
        )

    # Create static plot
    fig = debug_plot(models, input, target, figsize=figsize,
                     sampling_rate=sampling_rate, xlim=xlim)

    return fig, models


def animated_debug(model, input, target, iterations=None, sampling_rate=None,
                   xlim=None, figsize=None, save_path=None, load_path=None,
                   animation_save_path=None, frame_interval=500, backend='scipy',
                   fitter_options=None, **fit_kwargs):
    """Animate error, gradient, parameters, and prediction for each fit iteration.

    Parameters
    ----------
    model : nems.Model.
        Initial model to base fits on.
    input : np.ndarray, dict, or DataSet.
        See `nems.models.base.Model.evaluate`.
    target : np.ndarray or dict.
        See `nems.models.base.Model.fit`.
    iterations : int; optional.
        Number of fit iterations to record. For `backend='scipy'` this is
        interpreted as `{'options': {'maxiter': iterations}}`. For `backend='tf'`
        this is interpreted as `epochs=iterations`.
    sampling_rate : float; optional.
        If given, label time-axis of some plots with real units (seconds
        by default). See `nems.visualization.tools.ax_bins_to_seconds`.
    xlim : tuple; optional.
        Positional arguments for: `ax.set_xlim(*xlim)` for prediction plot.
        If `sampling_rate` is not None, specified in units of seconds instead
        of bins.
    figsize : 2-tuple of int; optional.
         Figure size in inches, (width, height).
    save_path : str or Path; optional.
        Path where models should be saved after each fit iteration. Must point
        to a directory.
    load_path : str or Path; optional.
        Path to load previously fit models from. Must point to a directory
        containing NEMS models saved to json format.
    animation_save_path : str or Path; optional.
        Path where animation should be saved. Rendering will be handled by
        `matplotlib.animation.FFMpegWriter`.
    frame_interval : int; default=500.
        Delay between frames in milliseconds.
    backend : str; default='scipy'.
        Name of backend to use for fitting. See `nems.models.base.Model.fit`.
    fitter_options : dict; optional.
        See `nems.models.base.Model.fit` and backend documentation.
    fit_kwargs : dict; optional.
        Additional keyword arguments for `nems.models.base.Model.fit`.

    Returns
    -------
    matplotlib.pyplot.figure
    list of nems.Model
        Length iterations + 1.

    See also
    --------
    matplotlib.animation.FuncAnimation
    matplotlib.animation.FFMpegWriter

    Notes
    -----
    Matplotlib animation seems to be very dependent on backend, IDE, OS, etc.
    For example, the animation will likely not play in a jupyter notebook. One
    work-around is to save the animation as a movie to watch separately, which
    is what the `animation_save_path` option is for. However, the FFmpeg encoder
    may or may not already be available on your system. If it is not, try
    `conda install -c conda-forge ffmpeg` (or use your preferred method).

    """
    
    # Get list of models, starting with initial model and appending one per
    # fit iteration.
    models = get_model_fits(
        model, input, target, iterations=iterations, backend=backend,
        fitter_options=fitter_options, save_path=save_path, load_path=load_path,
        **fit_kwargs
        )

    # Start with full figure to set limits and color scales.
    fig = debug_plot(
        models, input, target, figsize=figsize, sampling_rate=sampling_rate,
        xlim=xlim
        )

    # Get references to figure artists and their x/ydata for each iteration.
    artists, xdata, ydata = _get_animation_data(fig, input, models, iterations)

    # On each iteration, update x/ydata but leave all axes, labels, etc in place.
    def animation_step(i):
        for j, a in enumerate(artists):
            if isinstance(a, matplotlib.lines.Line2D):
                a.set_xdata(xdata[i][j])
                a.set_ydata(ydata[i][j])
            else:
                a.set_offsets(xdata[i][j])

        return artists

    anim = animation.FuncAnimation(
        fig, animation_step, frames=iterations, interval=frame_interval,
        blit=True
    )

    if animation_save_path is not None:
        fps = int(np.ceil(1/(frame_interval/1000)))
        # NOTE: You may need to install the FFmpeg encoder, for example using
        #       `conda install -c conda-forge ffmpeg`.
        writer = animation.FFMpegWriter(
            fps=fps, metadata=dict(artist='Me'), bitrate=1800
            )
        anim.save(animation_save_path, writer=writer)

    return anim, models


def _get_animation_data(fig, input, models, iterations):
    """Get matplotlib artists and data. Internal for `animated_debug`."""
    # Extract artists for initial figure, and get list of artist data for each
    # subsequent iteration.
    artists = []
    for i, ax in enumerate(fig.axes):
        artists.extend(ax.lines)
        if i == 2:
            # 3rd subplot is PCA scatter, have to add scatter points separately
            # (index 1). Index 0 is the line, 2 and 3 are initial/final.
            artists.append(ax.get_children()[1])

    updated_xdata = []
    updated_ydata = []
    for i in range(2, iterations+2):
        updated_xdata.append([])
        updated_ydata.append([])
        for j, a in enumerate(artists):
            # TODO: change this if more plots added, and/or refactor to not need
            #       a "magic numbers" here
            # TODO: this will break for TF b/c one line plot will be missing
            if j <= 4:
                if isinstance(a, matplotlib.lines.Line2D):
                    # iterations on x-axis
                    new_x = a.get_xdata()[:i]
                    new_y = a.get_ydata()[:i]
                else:
                    # scatter plot PathCollection, only one set of data.
                    new_x = a.get_offsets().data[:i]
                    new_y = []

            elif j == 5:
                # Actual response doesn't change, input heatmap isn't in artists.
                new_x = a.get_xdata()
                new_y = a.get_ydata()
            else:
                # Get new prediction, same x-values
                new_x = a.get_xdata()
                new_y = models[i-1].predict(input)

            updated_xdata[i-2].append(new_x)
            updated_ydata[i-2].append(new_y)

    return artists, updated_xdata, updated_ydata


# TODO: multiple initial conditions instead of iterations.



import numpy as np

from nems.registry import layer
from .base import Layer, Phi, Parameter


class SwapDims(Layer):
    """Swap two dimensions of an input array.

    See also
    --------
    nems.layers.base.Layer

    Examples
    --------
    >>> sd = SwapDims(dim1=1, dim2=2)
    >>> input = np.random.rand(10000, 18, 2)  # (time, channels, bands)
    >>> out = np.moveaxis(input, self.dim1, self.dim2)      # sd.evaluate(input)
    >>> out.shape
    (10000, 2, 18)

    """
 
    def __init__(self, dim1=1, dim2=2, **kwargs):
        self.dim1 = dim1
        self.dim2 = dim2
        super().__init__(**kwargs)

    def evaluate(self, input):
        """Swap two dimensions of the input."""
        return np.moveaxis(input, [self.dim1, self.dim2], [self.dim2, self.dim1])

    @layer('sd')
    def from_keyword(keyword):
        """Construct SwapDims from keyword.

        Keyword options
        ---------------
        {digit} : First dimension to swap; optional, default=1.
        {digit} : Second dimension to swap; optional, default=2.

        See also
        --------
        Layer.from_keyword
        
        """

        options = keyword.split('.')[1:3]
        kwargs = {f'dim{i+1}': d for i, d in enumerate(options)}
        return SwapDims(**kwargs)

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs."""
        
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer
  
        dim1 = self.dim1 + 1
        dim2 = self.dim2 + 1
    
        class SwapDimsTF(NemsKerasLayer):
            @tf.function
            def call(self, inputs):
                out = tf.experimental.numpy.moveaxis(
                    inputs, [dim1, dim2], [dim2, dim1]
                    )
                return out

        return SwapDimsTF(self, **kwargs)

    @property
    def plot_options(self):
        """Add legend at right of plot, with default formatting.

        Notes
        -----
        The legend will grow quite large if there are many output channels,
        but for common use cases (< 10) this should not be an issue. If needed,
        increase figsize to accomodate the labels.

        See also
        --------
        Layer.plot
        
        """
        return {'legend': True}

